#include "queue.h"
#include "stdio.h"
#include "termio.h"
#include "mips.h"

#define CNTRL(x) ('x'&0x1f)

#ifdef TEST
# define read(x,y,z) xread(x,y,z)
# define write(x,y,z) xwrite(x,y,z)
# define ioctl(x,y,z) xioctl(x,y,z)
  int SIOBASE[16];
#endif

typedef struct DevEntry {
	int txoff;
	int qsize;
	Queue *rxq;
	Addr sio;
	int chan;
	int rxoff;
	iFunc *handler;
	jmp_buf *intr;
	char *tname;
	iFunc *tfunc;
	struct termio t;
	} DevEntry;

DevEntry DevTable[DEV_MAX];

typedef struct File {
	int dev;
	int valid;
	} File;

File _file[OPEN_MAX] = {
	{0,1},{0,1},{0,1}
	};

struct TermEntry {
	char *name;
	iFunc *func;
	};

int tvi920(),vt100();

struct TermEntry TermTable[] = {
	{"tvi920",tvi920},
	{"vt100",vt100},
	{0}};

#ifdef TEST
int fd;

main()
{

devinit();
fd = 0;
for (;;) {
	if (!docmd("main>")) break;
	}
}

docmd(s)
char *s;
{
char buf[40];
struct termio t;
int i,value,reg;
DevEntry *p;

printf("fd=%d ",fd);
p = &DevTable[_file[fd].dev];
printf("txoff=%d rxq=",p->txoff);
prqueue(p->rxq);
for (i=0;i<16;i++) printf("%02x ",SIOBASE[i]);
printf("\n");
printf("pgfscq.%s",s);
gets(buf);
if (buf[0] == 0) return(1);
switch (buf[0]) {
	case 'q' :
		return(0);
		break;
	case 'p' :
		sscanf(buf,"p %x",&value);
		buf[0] = value;
		write(fd,buf,1);
		break;
	case 'G' :
		value = read(fd,buf,40);
		for (i=0;i<value;i++) printf("%02x ",buf[i]);
		printf("\n");
		break;
	case 'g' :
		read(fd,buf,1);
		printf("%02x\n",buf[0]);
		break;
	case 'f' :
		sscanf(buf,"f %d",&fd);
		break;
	case 'c' :
		ioctl(fd,TCGETA,&t);
		printf("iflag=%04x oflag=%04x cflag=%04x lflag=%04x\n",
			t.c_iflag,t.c_oflag,t.c_cflag,t.c_lflag);
		if (strlen(buf) == 1) break;
		sscanf(buf,"c %1s %04x",buf,&value);
		if (buf[0] == 'i') t.c_iflag = value;
		else if (buf[0] == 'o') t.c_oflag = value;
		else if (buf[0] == 'c') t.c_cflag = value;
		else if (buf[0] == 'l') t.c_lflag = value;
		printf("iflag=%04x oflag=%04x cflag=%04x lflag=%04x\n",
			t.c_iflag,t.c_oflag,t.c_cflag,t.c_lflag);
		ioctl(fd,TCSETAF,&t);
		break;
	case 's' :
		sscanf(buf,"s %d %x",&reg,&value);
		SIOBASE[reg] = value;
		break;
	}
return(1);
}

prqueue(q)
Queue *q;
{
int i,ch;

printf("[");
i = 0;
while ((ch = Qread(q,i)) != 0) {
	printf("%02x ",ch);
	i++;
	}
printf("]\n");
}
#endif

/** read(fd,buf,n) read n bytes into buf from fd */
read(fd,buf,n)
int fd,n;
char *buf;
{
int i,used;
DevEntry *p;
char ch;

p = &DevTable[_file[fd].dev];
for (i=0;i<n;) {
	scandevs();
	while ((used=Qinquiry(p->rxq,Q_USED)) == 0) reschedule("read Qempty");
	if (used < 10 && p->rxoff) {
		p->rxoff = 0;
		_write(p,CNTRL(Q));
		}
	ch = Qget(p->rxq);
	if (p->t.c_iflag&ICRNL && ch == '\r') ch = '\n';
	if (p->t.c_lflag&ECHO) write(fd,&ch,1);
	if (p->t.c_lflag&ICANON) {
		if (ch == p->t.c_cc[VEOL]) break;
		else if (ch == p->t.c_cc[VEOL2]) break;
		else if (ch == p->t.c_cc[VERASE]) {
			if (i > 0) i--;
			}
		else buf[i++] = ch;
		}
	else buf[i++] = ch;
	}
return(i);
}

/** write(fd,buf,n) write n bytes from buf to fd */
write(fd,buf,n)
int fd,n;
char *buf;
{
int i;
DevEntry *p;
char *t;

p = &DevTable[_file[fd].dev];
for (i=0;i<n;i++) {
	scandevs();
	if (p->t.c_oflag&ONLCR && buf[i] == '\n') _write(p,'\r');
	if (p->t.c_lflag&ECHOE && buf[i] == p->t.c_cc[VERASE]) {
		for (t="\b \b";*t;t++) _write(p,*t);
		}
	else _write(p,buf[i]);
	}
return(i);
}

static _write(p,ch)
DevEntry *p;
char ch;
{

while (p->txoff) reschedule("write txoff");
while (!(*p->handler)(OP_TXRDY,p->sio,p->chan)) reschedule("write txrdy");
(*p->handler)(OP_TX,p->sio,p->chan,ch);
}

static scandevs()
{
int i,c,n;
DevEntry *p;

for (i=0;DevTable[i].rxq;i++) {
	p = &DevTable[i];
	if ((*p->handler)(OP_RXRDY,p->sio,p->chan)) {
		c = (*p->handler)(OP_RX,p->sio,p->chan);
		if (p->t.c_iflag&ISTRIP) c &= 0x7f;
		n = Qinquiry(p->rxq,Q_SPACE);
		if (c == p->t.c_cc[VINTR] && p->intr) longjmp(p->intr,1);
		else if (p->t.c_iflag&IXON && p->t.c_iflag&IXANY 
			&& p->txoff) p->txoff = 0;
		else if (p->t.c_iflag&IXON && c == p->t.c_cc[V_STOP]) 
			p->txoff = 1;
		else if (p->t.c_iflag&IXON && c == p->t.c_cc[V_START]) 
			p->txoff = 0;
		else if (n > 0) {
			Qput(p->rxq,c);
			if (n < 10 && p->t.c_iflag&IXOFF && !p->rxoff) {
				p->rxoff = 1;
				_write(p,CNTRL(S));
				}
			}
		}
	}
}

/** open(fname,mode,perms) return fd for fname */
open(fname,mode,perms)
char *fname;
int mode,perms;
{
int i,c,dev,len;
char tmp[80];

for (i=0;i<OPEN_MAX && _file[i].valid;i++) ;
if (i == OPEN_MAX) return(-1);
len = strlen(fname);
if (len == 9) fname += 5;
else if (len != 4) return(-1);
strncpy(tmp,fname,3);
if (!strequ(tmp,"tty")) return(-1);
c = fname[3];
if (c >= 'a' && c <= 'z') dev = c - 'a';
if (c >= 'A' && c <= 'Z') dev = c - 'A';
if (c >= '0' && c <= '9') dev = c - '0';
if (dev > DEV_MAX) return(-1);
if (DevTable[dev].rxq == 0) return(-1);
_file[i].valid = 1;
_file[i].dev = dev;
ioctl(i,SETSANE);
return(i);
}

/** close(fd) close fd */
close(fd)
int fd;
{
_file[fd].valid = 0;
}

/** ioctl(fd,op,argp) perform control operation on fd */
ioctl(fd,op,argp)
int fd,op,*argp;
{
DevEntry *p;
struct termio *t;
int i;

switch (op) {
	case TCGETA :
		p = &DevTable[_file[fd].dev];
		t = (struct termio *) argp;
		*t = p->t;
		break;
	case TCSETAF : /* after flush of input queue */
		p = &DevTable[_file[fd].dev];
		while (! Qempty(p->rxq)) Qget(p->rxq);
	case TCSETAW : /* after write */
		/* no txq, so no delay needed */
		p = &DevTable[_file[fd].dev];
		if (p->t.c_cflag != ((struct termio *)argp)->c_cflag) {
			if ((*p->handler)(OP_BAUD,p->sio,p->chan,
				CBAUD&((struct termio *)argp)->c_cflag)) 
					return(-1);
			}
		p->t = *((struct termio *) argp);
		break;
	case FIONREAD :
		p = &DevTable[_file[fd].dev];
		scandevs();
		*argp = Qinquiry(p->rxq,Q_USED);
		break;
	case SETINTR :
		p = &DevTable[_file[fd].dev];
		p->intr = (jmp_buf *)argp;
		break;
	case GETINTR :
		p = &DevTable[_file[fd].dev];
		*argp = (unsigned long)p->intr;
		break;
	case SETSANE :
		p = &DevTable[_file[fd].dev];
		p->t.c_iflag |= (ISTRIP|ICRNL|IXON);
		p->t.c_oflag = (ONLCR);
		p->t.c_lflag = (ICANON|ECHO|ECHOE);
		p->t.c_cc[VINTR] = CNTRL(c);
		p->t.c_cc[VEOL] = '\n';
		p->t.c_cc[VEOL2] = CNTRL(c);
		p->t.c_cc[VERASE] = CNTRL(h);
		p->t.c_cc[V_STOP] = CNTRL(s);
		p->t.c_cc[V_START] = CNTRL(q);
		break;
	case SETNCNE : 
		p = &DevTable[_file[fd].dev];
		if (argp != 0) {
			t = (struct termio *) argp;
			*t = p->t;
			}
		p->t.c_lflag &= ~(ICANON|ECHO|ECHOE);
		p->t.c_cc[4] = 1;
		break;
	case CBREAK :
		p = &DevTable[_file[fd].dev];
		if (argp != 0) {
			t = (struct termio *) argp;
			*t = p->t;
			}
		p->t.c_lflag &= ~(ICANON|ECHO);
		p->t.c_cc[4] = 1;
		break;
	case GETTERM :
		p = &DevTable[_file[fd].dev];
		*argp = 0;
		if (p->tfunc == 0) return(-1);
		strcpy((char *)argp,p->tname);
		break;
	case SETTERM :
		p = &DevTable[_file[fd].dev];
		for (i=0;TermTable[i].name;i++) {
			if (strequ(argp,TermTable[i].name)) break;
			}
		if (TermTable[i].name == 0) return(-1);
		p->tname = TermTable[i].name;
		p->tfunc = TermTable[i].func;
		break;
	case TERMTYPE :
		if (TermTable[fd].name == 0) return(-1);
		strcpy((char *)argp,TermTable[fd].name);
		return(fd+1);
		break;
	default:
		return(-1);
	}
return(0);
}

extern ConfigEntry ConfigTable[];

devinit()
{
int i;
ConfigEntry *q;
DevEntry *p;

for (i=0;ConfigTable[i].devinfo && i < DEV_MAX;i++) {
	q = &ConfigTable[i];
	p = &DevTable[i];
	p->txoff = 0;
	p->rxoff = 0;
	if (q->chan == 0) (*q->handler)(OP_INIT,q->devinfo);
	p->qsize = q->rxqsize;
	p->rxq = Qcreate(p->qsize);
	if (p->rxq == 0) return(-1);
	(*q->handler)(OP_BAUD,q->devinfo,q->chan,q->brate);
	p->sio = q->devinfo;
	p->chan = q->chan;
	p->handler = q->handler;
	p->intr = 0;
	p->tfunc = 0;
	p->t.c_iflag = (ISTRIP|ICRNL|IXON);
	p->t.c_oflag = (ONLCR);
	p->t.c_lflag = (ICANON|ECHO|ECHOE);
	p->t.c_cflag = q->brate;
	p->t.c_cc[VINTR] = CNTRL(c);
	p->t.c_cc[VEOL] = '\n';
	p->t.c_cc[VEOL2] = CNTRL(c);
	p->t.c_cc[VERASE] = CNTRL(h);
	p->t.c_cc[V_STOP] = CNTRL(s);
	p->t.c_cc[V_START] = CNTRL(q);
	_write(p,CNTRL(Q));
	}
return(0);
}

reschedule(p)
char *p;
{ 
#ifdef TEST
char buf[80];

scandevs();
sprintf(buf,"%s>>",p);
docmd(buf);
#else
scandevs();
#endif
}

/** ttctl(fd,op,a1,a2) perform terminal specific operation */
ttctl(fd,op,a1,a2)
int fd,op,a1,a2;
{
DevEntry *p;
int r;

p = &DevTable[_file[fd].dev];
if (p->tfunc == 0) return(-1);
r = (*p->tfunc)(fd,op,a1,a2);
return(r);
}

