#include "mips.h"

#define CLKRATE  40	/* CPU clock in nanosecs */
#define INTRATE 100	/* interrupt rate in millisecs */
#define CLKVAL ((INTRATE*1000000)/40)
#define SCC_BASE 0xbe000000
#define LED_BIT 0x01

	.comm flag,4
	.comm _time,4
	.comm tenths,4

	.globl clkisr
	.ent clkisr
	.set noat
clkisr: # k1 = rtn addr to exception handler
	# first check to see if it's an int0
	.set noreorder
	mfc0	k0,C0_CAUSE
	nop
	.set reorder
	# if it's not for me, return
	sll	k0,21
	bltz	k0,1f	# branch if bit 10 is set (int0)
	j	k1	# return to exception handler

   # It is an int0 so first acknowledge the timer's int request
1: # do this by clearing the INT bit (bit 0).	
	li	k0,TC1
	lw	k1,(k0)
	srl	k1,1
	sll	k1,1
	sw	k1,(k0)

	# inc tenths
	la	k0,tenths
	lw	k1,(k0)
	addu	k1,1
	sw	k1,(k0)

	# overflow?
	subu	k1,10
	bne	k1,zero,2f

	# yes, clr tenths
	la	k0,tenths
	sw	zero,(k0)

	# update time
	la	k0,_time
	lw	k1,(k0)
	addu	k1,1
	sw	k1,(k0)

   # To flash the LED first we must find it's present state
	la	k0,flag
	lw	k1,(k0)
	beq	k1,zero,1f	# is it on?
	move	k1,zero		# turn it off
	sw	k1,(k0)		# store the new state

   # turn the LED off
	li	k0,SCC_BASE
	li	k1,LED_BIT
	sb	k1,0x3b(k0)
	b	2f

1:	li	k1,1		# turn it on
	sw	k1,(k0)		# store the new state

   # turn the LED on
	li	k0,SCC_BASE
	li	k1,LED_BIT
	sb	k1,0x3f(k0)

2: # return to the interrupted program
	.set noreorder
	mfc0	k0,C0_EPC
	nop
	j	k0
	rfe
	.set reorder
	.set at
	.end clkisr

	.globl stime
	.ent stime
stime:
	la	t0,_time
	sw	a0,(t0)
	la	k0,tenths
	sw	zero,(k0)
	j	ra
	.end stime

	.globl time
	.ent time
time:
	la	t0,_time
	lw	v0,(t0)
	beq	a0,zero,1f
	sw	v0,(a0)
1:	j	ra
	.end time

	.globl clock
	.ent clock
clock:    # a0 = ptr to struct of three ints
	la	t1,_time
	lw	t0,(t1)
	sw	t0,0(a0)

	la	t1,tenths
	lw	t0,(t1)
	sw	t0,4(a0)

	la	t1,TIC1
	lw	t0,(t1)
	li	t1,CLKVAL
	subu	t1,t0
	mul	t0,t1,40
	sw	t0,8(a0)
	j	ra
	.end clock


	.globl startclk
	.ent startclk
startclk: 
	# enable int0 and IEC
	li	t1,(SR_IEC|SR_IBIT3)
	.set noreorder
	mtc0	t1,C0_SR
	.set reorder

	# set the rate
	li	t1,TIC1
	li	t0,CLKVAL
	sw	t0,(t1)

	# enable counting and ints
	li	t0,(TCTL_CE|TCTL_IE)
	li	t1,TC1
	sw	t0,(t1)
	j	ra
	.end startclk

