#include "mips.h"
#include "pmon.h"
#include "stdio.h"
#include "termio.h"

#define comma()		strcat(dest,",")
#define rd()		strcat(dest,regname[(int)RD_(inst)])
#define rs()		strcat(dest,regname[(int)RS_(inst)])
#define rt()		strcat(dest,regname[(int)RT(inst)])
#define c0()		strcat(dest,c0reg[(int)RD_(inst)])
#define c1()		strcat(dest,c1reg[(int)RD_(inst)])
#define cn()		strcat(dest,regs_hw[(int)RD_(inst)])
#define c0ft()		strcat(dest,c0reg[(int)RT(inst)])
#define c1ft()		strcat(dest,c1reg[(int)RT(inst)])
#define cnft()		strcat(dest,regs_hw[(int)RT(inst)])
#define cc1()		strcat(dest,regs_hw[(int)RD_(inst)])

long inst, cur_loc;


char **regname;

char *regs_sw[] = {
	"zero",  "at", "v0",  "v1",  "a0",  "a1",  "a2",  "a3",
	"t0",  "t1",  "t2",  "t3",  "t4",  "t5",  "t6",  "t7",
	"s0",  "s1",  "s2",  "s3",  "s4",  "s5",  "s6",  "s7",
	"t8",  "t9",  "k0",  "k1",  "gp",  "sp",  "s8",  "ra"
	};

char *regs_hw[] = {
	"$0",  "$1",  "$2",  "$3",  "$4",  "$5",  "$6",  "$7",
	"$8",  "$9",  "$10", "$11", "$12", "$13", "$14", "$15",
	"$16", "$17", "$18", "$19", "$20", "$21", "$22", "$23",
	"$24", "$25", "$26", "$27", "$28", "$29", "$30", "$31"
	};

char **c0reg;

char *regs_c0[] = {
	"C0_INX",     "C0_RAND",  "C0_TLBLO",  "C0_BPC",  
	"C0_CTEXT",   "C0_BDA",   "$6",        "C0_DCIC",
 	"C0_BADADDR", "$9",       "C0_TLBHI",  "$11", 
	"C0_SR",      "C0_CAUSE", "C0_EPC",    "C0_PRID",
	"$16", "$17", "$18", "$19", "$20", "$21", "$22", "$23",
	"$24", "$25", "$26", "$27", "$28", "$29", "$30", "$31"
	};

char *c1reg[] = {
	"$f0",  "$f1",  "$f2",  "$f3",  "$f4",  "$f5",  "$f6",  "$f7",
	"$f8",  "$f9",  "$f10", "$f11", "$f12", "$f13", "$f14", "$f15",
	"$f16", "$f17", "$f18", "$f19", "$f20", "$f21", "$f22", "$f23",
	"$f24", "$f25", "$f26", "$f27", "$f28", "$f29", "$f30", "$f31"
	};

long disasm(char *dest,long loc, long bits){
	DISTBL *pt;
	char *pc;
	long v,v1,w;
	int i;
	char tmp[40];

	cur_loc = loc;
	inst = bits;

	if(!adr2symoff(dest,cur_loc,12)) sprintf(dest,"%08x",cur_loc);
	sprintf(tmp," %08x ",inst);
	strcat(dest,tmp);

	if(inst == 0L){
		strcat(dest,"nop");
		return(loc + 4L);
	}
	pt = get_distbl(inst);
	i = strlen(pt->str);
	strcat(dest,pt->str);
	while(i++ < 8) strcat(dest," ");
	switch(pt->type) {
	case RT_RS_IMM:	rt(), comma(), rs(), comma(),imm(dest);	break;
	case RT_RS_SIMM:rt(), comma(), rs(),comma(),simm(dest);	break;
	case RT_IMM:	rt(), comma(), imm(dest);		break;
	case RT_SIMM:	rt(), comma(), simm(dest);		break;
	case RT_RD:	rt(), comma();
	case RD:	rd();					break;
	case RT_C0:	rt(),comma(),c0();			break;
	case RT_C1:	rt(),comma(),c1();			break;
	case RT_CN:	rt(),comma(),cn();			break;
	case RT_CC1:	rt(),comma(),cc1();			break;
	case RD_RT_RS:	rd(), comma(), rt(), comma();
	case JR: case RS:
			rs();					break;
	case RD_RS_RT:	rd(), comma();
	case RS_RT:	rs(), comma(), rt();			break;
	case RD_RT_SFT:	rd(), comma(), rt(), comma();
			sprintf(tmp,"0x%x",SHAMT(inst));
			strcat(dest,tmp);
			break;
	case RS_RT_OFF:
	case RS_OFF:
			rs(), comma();
			if(pt->type == RS_RT_OFF)
				rt(), comma();
	case OFF:
			v = IMM(inst);
			if(v & (1L << 15)) v |= 0xffff0000L;
			v1 = cur_loc + 4L + (v << 2);
			if (!adr2symoff(tmp,v1,0)) sprintf(tmp,"%x",v1);
			strcat(dest,tmp);
			break;
	case COFUN:	sprintf(tmp,"%x",inst & 0x01ffffffL);
			strcat(dest,tmp);
			break;
	case NONE:						break;
	case TARGET:
			v = (inst & 0x03ffffffL) << 2;
			v |= (cur_loc & 0xf0000000L);
			if (!adr2symoff(tmp,v,0)) sprintf(tmp,"%x",v);
			strcat(dest,tmp);
			break;
	case JALR:
			if(RD_(inst) != 31L)
				rd(), comma();
			rs();
			break;
	case LDSTC0:
			v = IMM(inst);
			if(v & (1L << 15)) v |= 0xffff0000L;
			c0ft(), comma();
			sprintf(tmp,"%d",v);
			strcat(dest,tmp);
			strcat(dest,"(");
			rs();
			strcat(dest,")");
			break;
	case LDSTC1:
			v = IMM(inst);
			if(v & (1L << 15)) v |= 0xffff0000L;
			c1ft(), comma();
			sprintf(tmp,"%d",v);
			strcat(dest,tmp);
			strcat(dest,"(");
			rs();
			strcat(dest,")");
			break;
	case LDSTCN:
			v = IMM(inst);
			if(v & (1L << 15)) v |= 0xffff0000L;
			cnft(), comma();
			sprintf(tmp,"%d",v);
			strcat(dest,tmp);
			strcat(dest,"(");
			rs();
			strcat(dest,")");
			break;
	case LOAD_STORE:
			v = IMM(inst);
			if(v & (1L << 15)) v |= 0xffff0000L;
			rt(), comma();
			sprintf(tmp,"%d",v);
			strcat(dest,tmp);
			strcat(dest,"(");
			rs();
			strcat(dest,")");
			break;
	case WORD:
			sprintf(tmp,"%08x",inst);
			strcat(dest,tmp);
			strcat(dest,"      # ");
			w = cur_loc;
			for(i = 0; i < 4 ; i++){
				v = load_byte(w++);
				if (isprint(v)) strccat(dest,v);
				else strccat(dest,'.');
				}
			break;
	}
	return(loc + 4L);
}

dis(ac,av)
int ac;
char *av[];
{
word adr = cur_adr, siz, i, l, ol;
byte v;
char prnbuf[LINESZ];


if(ac >= 2) {
	if(!get_rsa(&adr,av[1])) return(1);
	}
else adr = Epc;

if(ac == 3) {
	if(!get_rsa(&siz,av[2])) return(1);
	}
else {
	if (!atob(&siz,getenv("moresz"),10)) {
		printf("%s: bad moresz value\n",getenv("moresz"));
		return(1);
		}
	}


if (matchenv("regstyle")) {
	regname = regs_sw;
	c0reg = regs_c0;
	}
else {
	regname = regs_hw;
	c0reg = regs_hw;
	}

ioctl(STDIN,CBREAK,NULL);
l = siz;
while(1) {
	if (adr2sym(prnbuf,adr)) {
		strcat(prnbuf,":");
		if (more(prnbuf,&l,(ac>2)?0:siz)) break;
		}
	adr = disasm(prnbuf,adr, load_word(adr));
	if (more(prnbuf,&l,(ac>2)?0:siz)) break;
	}
}

void simm(char *dest){
char tmp[20];
long	v = IMM(inst);

if(v & (1L << 15)) v |= 0xffff0000L;
if (v >= 0) sprintf(tmp,"0x%x",v);
else sprintf(tmp,"-0x%x",v);
strcat(dest,tmp);
}

void imm(char *dest){
char tmp[20];
long	v = IMM(inst);

if (v >= 0) sprintf(tmp,"0x%x",v);
else sprintf(tmp,"-0x%x",v);
strcat(dest,tmp);
}

