#include "mips.h"
#include "pmon.h"
#include "stdio.h"
#include "termio.h"
#include "set.h"
#include "string.h"

byte tracing_mode;

/** go(ac,av) the go command */
go(ac,av)
int ac;
char *av[];
{
word	adr;
int i,had_epc,sflag;

had_epc = 0;
sflag = 0;
Tmpbpa = NO_BP;
Trcbp1a = NO_BP;
Trcbp2a = NO_BP;

Gpr[4] = 0; /* clear client's a0 (argc) */
for (i=1;i<ac;i++) {
	if (av[i][0] == '-') {
		if (av[i][1] == 's') sflag = 1;
		else if (av[i][1] == 'c') {
			Gpr[4] = ac-i;
			Gpr[5] = (word)&av[i];
			break;
			}
		else printf("%s: unknown option\n",av[i]);
		}
	else {
		if (had_epc) {
			if(!get_rsa(&adr,av[i])) return(1);
			Tmpbpa = adr;
			}
		else {
			had_epc++;
			if(!get_rsa(&adr,av[i])) return(1);
			Epc = adr;
			}
		}
	}
if (!sflag) Gpr[29] = clienttos();
goclient();
}

/** cont(ac,av) the continue command */
cont(ac,av)
int ac;
char *av[];
{
word	adr;

Tmpbpa = NO_BP;
Trcbp1a = NO_BP;
Trcbp2a = NO_BP;
if(ac > 1){
	if(!get_rsa(&adr,av[1])) return(1);
	Tmpbpa = adr;
	}
goclient();
}

/** trace(ac,av) the single-step command */
trace(ac,av)
int ac;
char *av[];
{
int stepover;

if (strequ(av[0],"to")) stepover = 1;
else stepover = 0;

Trcbp1a = NO_BP;
Trcbp2a = NO_BP;
if(is_branch(Epc)){
	if(is_branch(Epc + 4)){
		printf("branch in delay slot\n");
		return(1);
		}
	if(branch_target_address(Epc) == Epc){
		printf("self-branches can't trace\n");
		return(1);
		}		
	if(is_conditional_branch(Epc)) Trcbp1a = Epc + 8;
	if (is_jal(Epc) && stepover) Trcbp1a = Epc + 8;
	else Trcbp2a = branch_target_address(Epc);
	}
else Trcbp1a = Epc + 4;

tracing_mode = TRACE_TB;
store_trace_breakpoint();
_go();
}

/** setbp(ac,av) the set breakpoint command */
setbp(ac,av)
int ac;
char *av[];
{
word adr, i, j,w,x;
char *p;

if(ac > 1) {
	for(i = j = 0; j < ac - 1; j++) {
#ifdef LR33000
		if (av[1 + j][0] == '-') {
			p = &av[1 + j][1];
			w = (DCIC_TR|DCIC_UD|DCIC_KD|DCIC_DE);
			x = 0;
			for (;*p;p++) {
				switch (*p) {
					case 'i' : w |= DCIC_PCE; break;
					case 'd' : 
						w |= (DCIC_DAE|DCIC_DW|DCIC_DR);
						break;
					case 'k' : x |= DCIC_UD; break;
					case 'u' : x |= DCIC_KD; break;
					case 'r' : x |= DCIC_DW; break;
					case 'w' : x |= DCIC_DR; break;
					}
				}
			j++; /* get next arg */
			if(!get_rsa(&adr,av[1 + j])) w = 0;
			if (w&DCIC_PCE)  {
				Dcic |= w;
				Dcic &= ~x;
				Bpc = adr;
				printf("Bpt %2d = %08x\n",MAX_BP,adr);
				}
			else if (w&DCIC_DAE) {
				Dcic |= w;
				Dcic &= ~x;
				Bda = adr;
				printf("Bpt %2d = %08x\n",MAX_BP+1,adr);
				}
			else printf("error: breakpoint not set\n");
			}
		else if(get_rsa(&adr,av[1 + j])) {
#else
		if(get_rsa(&adr,av[1 + j])) {
#endif
			while(i < MAX_BP && Bpa[i] != NO_BP) i++;
			if(MAX_BP <= i) {
				printf("too many breakpoints\n");
				return(1);
				}
			else if(Bpa[i] != adr) {
				Bpa[i] = adr;
				printf("Bpt %2d = %08x",i,adr);
				if(adr & 3L) printf(" -> ??");
				putchar(NL);
				}
			}
		}
	}
else dspbpts();
return(0);
}

/** dspbpts() display all breakpoints */
dspbpts()
{
int i,ln,siz;
char tmp[64],buf[100];

if (!atob(&siz,getenv("moresz"),10)) {
	printf("%s: bad moresz value\n",getenv("moresz"));
	return(1);
	}

ioctl(STDIN,CBREAK,NULL);
ln = siz;
for(i = 0; i < MAX_BP; i++)
	if(Bpa[i] != NO_BP){
		sprintf(buf,"Bpt %2d = %08x ",i,Bpa[i]);
		if (adr2symoff(tmp,Bpa[i],0)) strcat(buf,tmp);
		if(Bpa[i] & 3L) strcat(buf," -> ??");
		if (more(buf,&ln,siz)) break;
		}

#ifdef LR33000
if (Dcic&DCIC_DE) {
	if (Dcic&DCIC_PCE) {
		printf("Bpt %2d = %08x",MAX_BP,Bpc);
		if (adr2symoff(tmp,Bpa[i],0)) printf(" %s",tmp);
		printf(" instr");
		if (Dcic&DCIC_KD) printf(" kernel");
		if (Dcic&DCIC_UD) printf(" user");
		if (adr2symoff(tmp,Bpc,0)) printf(" %s",tmp);
		printf("\n");
		}
	if (Dcic&DCIC_DAE) {
		printf("Bpt %2d = %08x",MAX_BP+1,Bda);
		if (adr2symoff(tmp,Bpa[i],0)) printf(" %s",tmp);
		printf(" data");
		if (Dcic&DCIC_KD) printf(" kernel");
		if (Dcic&DCIC_UD) printf(" user");
		if (Dcic&DCIC_DR) printf(" read");
		if (Dcic&DCIC_DW) printf(" write");
		if (adr2symoff(tmp,Bda,0)) printf(" %s",tmp);
		printf("\n");
		}
	}
#endif
}

clrbp(ac,av)
int ac;
char *av[];
{
word adr, i, j;

if(ac > 1){
	for(i = j = 0; j < ac - 1; j++){
		if (strequ(av[1+j],"*")) {
			clrbpts();
			continue;
			}
		if (!atob(&i,av[1+j],10)) {
			printf("%s: decimal number expected\n",av[1+j]);
			continue;
			}
		if (i < MAX_BP) {
			if (Bpa[i] == NO_BP) printf("Bpt %2d not set\n",i);
			else Bpa[i] = NO_BP;
			}
#ifdef LR33000
		else if (i == MAX_BP) {
			if ((Dcic&DCIC_PCE) && (Dcic&DCIC_DE)) {
				Dcic &= ~DCIC_PCE;
				if (!(Dcic&DCIC_DAE)) Dcic &= ~DCIC_DE;
				}
			else printf("Bpt %2d not set\n",i);
			}
		else if (i == MAX_BP+1) {
			if ((Dcic&DCIC_DAE) && (Dcic&DCIC_DE)) {
				Dcic &= ~DCIC_DAE;
				if (!(Dcic&DCIC_PCE)) Dcic &= ~DCIC_DE;
				}
			else printf("Bpt %2d not set\n",i);
			}
#endif
		else printf("%d: breakpoint number too large\n",i);
		}
	}
else dspbpts();
return(0);
}

save_bpts()
{
int i;

if (tracing_mode = TRACE_NO) return;

if(tracing_mode == TRACE_GB){
	if(Tmpbpa != NO_BP){
		save_bp(Tmpbpa, Tmpbpv);
		Tmpbpa = NO_BP;
		}
	for(i = 0; i < MAX_BP; i++)
		if(Bpa[i] != NO_BP)
			save_bp(Bpa[i], Bpv[i]);
	}
else if(Trcbp1a != NO_BP){
	save_bp(Trcbp1a, Trcbp1v);
	Trcbp1a = NO_BP;
	}
else if(Trcbp2a != NO_BP){
	save_bp(Trcbp2a, Trcbp2v);
	Trcbp2a = NO_BP;
	}
tracing_mode = TRACE_NO;
}

clrbpts()
{
int i;

for(i = 0; i < MAX_BP; i++) Bpa[i] = NO_BP;
#ifdef LR33000
Dcic &= ~DCIC_DE;
#endif
}

goclient()
{
word	adr;

if(is_break_point(Epc)){
	if(is_branch(Epc)){
		if(is_branch(Epc + 4)){
			printf("branch in delay slot\n");
			return(1);
			}
		if(branch_target_address(Epc) == Epc){
			printf("self-branches can't trace\n");
			return(1);
			}
		if(is_conditional_branch(Epc)) Trcbp1a = Epc + 8;
		Trcbp2a = branch_target_address(Epc);
		}
	else Trcbp1a = Epc + 4;
	tracing_mode = TRACE_TG;
	store_trace_breakpoint();
	}
else{
	tracing_mode = TRACE_GB;
	store_breakpoint();
	}
_go();
}

sstep()
{

Trcbp1a = NO_BP;
Trcbp2a = NO_BP;
if(is_branch(Epc)){
	if(is_branch(Epc + 4)){
		printf("branch in delay slot\n");
		return(1);
		}
	if(branch_target_address(Epc) == Epc){
		printf("self-branches can't trace\n");
		return(1);
		}		
	if(is_conditional_branch(Epc)) Trcbp1a = Epc + 8;
	Trcbp2a = branch_target_address(Epc);
	}
else Trcbp1a = Epc + 4;

tracing_mode = TRACE_DS;
store_trace_breakpoint();
_go();
}

void exception(void){
int	i;

if(tracing_mode == TRACE_NO){
	printf("\nexception!\n");
	reg_rdbg();
	}
else if(tracing_mode == TRACE_GB){
	if(Tmpbpa != NO_BP){
		save_bp(Tmpbpa, Tmpbpv);
		Tmpbpa = NO_BP;
		}
	for(i = 0; i < MAX_BP; i++)
		if(Bpa[i] != NO_BP)
			save_bp(Bpa[i], Bpv[i]);
	reg_rdbg();
	}
if(Trcbp1a != NO_BP){
	save_bp(Trcbp1a, Trcbp1v);
	Trcbp1a = NO_BP;
	}
if(Trcbp2a != NO_BP){
	save_bp(Trcbp2a, Trcbp2v);
	Trcbp2a = NO_BP;
	}
if(tracing_mode == TRACE_TB) reg_rdbg();
else if(tracing_mode == TRACE_DC) dbgmode(1);
else if(tracing_mode == TRACE_DS) dbgmode(2);

tracing_mode = TRACE_GB;
store_breakpoint();
_go();
}

void reg_rdbg(void)
{
char cmd[LINESZ];

tracing_mode = TRACE_NO;
strcpy(cmd,getenv("brkcmd"));
do_cmd(cmd);
rdbg();
}

void store_breakpoint(void){
	int	i;

	for(i = 0; i < MAX_BP; i++){
		if(Tmpbpa == Bpa[i])
			Tmpbpa = NO_BP;
		if(Trcbp1a == Bpa[i])
			Trcbp1a = NO_BP;
		if(Trcbp2a == Bpa[i])
			Trcbp2a = NO_BP;
	}
	if(Trcbp1a == Tmpbpa)
		Trcbp1a = NO_BP;
	if(Trcbp2a == Tmpbpa || Trcbp2a == Trcbp1a)
		Trcbp2a = NO_BP;

	for(i = 0; i < MAX_BP; i++)
		if(Bpa[i] != NO_BP){
			Bpv[i] = load_bp(Bpa[i]);
			save_bp(Bpa[i], BP_CODE);
		}
	if(Tmpbpa != NO_BP){
		Tmpbpv = load_bp(Tmpbpa);
		save_bp(Tmpbpa, BP_CODE);
	}
	store_trace_breakpoint();
}

void store_trace_breakpoint(void){
	if(Trcbp1a != NO_BP){
		Trcbp1v = load_bp(Trcbp1a);
		save_bp(Trcbp1a, BP_CODE);
	}
	if(Trcbp2a != NO_BP){
		Trcbp2v = load_bp(Trcbp2a);
		save_bp(Trcbp2a, BP_CODE);
	}
}

int is_break_point(word adr){
	int	i;

	for(i = 0; i < MAX_BP; i++)
		if(Bpa[i] == adr)
			return(1);
	if(Tmpbpa == adr)
		return(1);
	return(0);
}

