#include "mips.h"
#include "pmon.h"
#include "stdio.h"
#include "termio.h"
#include "set.h"
#include "string.h"

/*
 * Useful S-records for testing purposes:
 *	S0030000FC
 *	S309900200001234567860
 *	S7058002000073
 *
 * Useful fast-records for testing purposes:
 *	/ACgAgAA	- addr = a0020000
 *	/ZAI		- clear 24 bytes
 *	/E		- define entry at 'addr' and exit
 *	AAABAAAC	- 00 00 01 00 00 02
 *	/Sfred,X	- symbol = 'addr'
 *	/BAB		- byte of 01
 */

#define MSK 0xff
#define BLKSIZE 512

unsigned long nextaddr;
unsigned long wordbuf;
int bytecnt = 0;
int blkinx = 0;
int blksz = 0;
char curblk[BLKSIZE];

char exta = ',';
char extb = '.';
char extc = '/';

load (ac, av)
     int ac;
     char *av[];
{
  struct termio tbuf;
  char *hostport, *typestr, *cmdstr;
  int fd, dlecho, dlproto, dltype;
  int len, err, i, j, tot, sflag, bflag;

  cmdstr = 0;
  sflag = bflag = 0;
  for (i = 1; i < ac; i++)
    {
      if (av[i][0] == '-')
	{
	  j = 1;
	  while (av[i][j] != 0)
	    {
	      if (av[i][j] == 's')
		sflag = 1;
	      else if (av[i][j] == 'b')
		bflag = 1;
	      else if (av[i][j] == 'c')
		{
		  i++;
		  if (i >= ac)
		    {
		      printf ("bad arg count\n");
		      return;
		    }
		  cmdstr = av[i];
		  break;
		}
	      else
		printf ("%c: unrecognize option\n", av[i][j]);
	      j++;
	    }
	}
      else
	printf ("%s: unrecognized argument\n", av[i]);
    }

  if (!sflag)
    clrsyms ();
  if (!bflag)
    clrbpts ();

  hostport = getenv ("hostport");
  fd = open (hostport, 0, 0);
  if (fd == -1)
    {
      printf ("can't open %s\n", hostport);
      return;
    }

  dltype = matchenv ("dltype");
  if (dltype == 0)
    typestr = "S-records";
  else if (dltype == 1)
    typestr = "fast-format";
  else
    {
      printf ("dltype %s not recognized\n", getenv ("dltype"));
      return;
    }

  printf ("Downloading %s from %s, ^C to abort\n", typestr, hostport);
  ioctl (fd, TCGETA, &tbuf);
  tbuf.c_iflag &= ~IXOFF;
  ioctl (fd, TCSETAF, &tbuf);

  dlecho = matchenv ("dlecho");
  if (dlecho == OFF || dlecho == LFEED)
    {
      tbuf.c_lflag &= ~ECHO;
      ioctl (fd, TCSETAF, &tbuf);
    }

  dlproto = matchenv ("dlproto");

  if (cmdstr)
    {
      write (fd, cmdstr, strlen (cmdstr));
      write (fd, "\r", 1);
    }

  err = tot = 0;
  while (1)
    {
      if (dltype == 1)
	len = fload (fd, dlproto);
      else if (dltype == 0)
	len = sload (fd, dlproto);
      if (len == -20)
	break;

      if (len < 0)
	err++;
      else
	tot += len;

      if (dlproto == XONXOFF)
	writec (fd, CNTRL (q));
      else if (dlproto == ETXACK)
	writec (fd, CNTRL (f));
      if (dlecho == LFEED)
	writec (fd, '\n');
    }
  if (dlproto == XONXOFF)
    writec (fd, CNTRL (q));
  else if (dlproto == ETXACK)
    writec (fd, CNTRL (f));
  if (dlecho == LFEED)
    writec (fd, '\n');
  flush_cache (ICACHE);

  if (err)
    printf ("\n%d errors\n", err);
  else
    printf ("\ntotal = %08x bytes\n", tot);
}

sload (fd, dlproto)
     int fd, dlproto;
{
  word adr;
  byte *p;
  int len, i;
  char recbuf[MAXREC];

  read (fd, recbuf, MAXREC);
  if (dlproto == XONXOFF)
    writec (fd, CNTRL (s));
  len = 0;
  if (recbuf[0] == 'S')
    {
      if ('7' <= recbuf[1] && recbuf[1] <= '9')
	{
	  endrec (recbuf);
	  return (-20);
	}
      else if (recbuf[1] == '4')
	dlsym (recbuf);
      else if (0 < (len = sfmt_load (pattern, recbuf, &adr)))
	{
	  for (p = pattern, i = len; i-- > 0; adr++, p++)
	    *(byte *) adr = *p;
	}
    }
  return (len);
}

word 
sfmt_load (byte * dst, byte * src, word * adrp)
{
  byte fmt, val;
  word i, len;

  chksum = *adrp = 0;
  if (*src++ != 'S')
    return (-1);
  fmt = *src++;
  if (0 == (src = scan_byte (&val, src)))
    return (-2);
  len = val;
  switch (fmt)
    {
    case '0':
      return (0);
    case '7':
    case '8':
    case '9':
      return (-20);
    case '3':
      if (0 == (src = scan_byte (&val, src)))
	return (-3);
      *adrp |= (val & 0xff) << 24;
      /* fall */
    case '2':
      if (0 == (src = scan_byte (&val, src)))
	return (-4);
      *adrp |= (val & 0xff) << 16;
      /* fall */
    case '1':
      if (0 == (src = scan_byte (&val, src)))
	return (-5);
      *adrp |= (val & 0xff) << 8;
      if (0 == (src = scan_byte (&val, src)))
	return (-6);
      *adrp |= (val & 0xff) << 0;
      break;
    default:
      return (-7);
    }
  len -= 3 + (fmt - '1');
  for (i = 0; i < len; i++)
    {
      if (0 == (src = scan_byte (&val, src)))
	return (-8);
      *dst++ = val;
    }
  if (0 == (src = scan_byte (&val, src)))
    return (-9);
  if (~chksum & 0xff)
    return (-10);
  return (len);
}

byte *
scan_byte (byte * valp, byte * src)
{
  word i;

  for (i = 2, *valp = 0; i-- > 0; src++)
    {
      *valp << = 4;
      if ('0' <= *src && *src <= '9')
	*valp += *src - '0';
      else if ('A' <= *src && *src <= 'F')
	*valp += *src - 'A' + 10;
      else if ('a' <= *src && *src <= 'f')	/* allow lowercase */
	*valp += *src - 'f' + 10;
      else
	return (0);
    }
  chksum += *valp;
  return (src);
}

writec (fd, ch)
     int fd;
     char ch;
{
  write (fd, &ch, 1);
}

dlsym (p)
     char *p;
{
  byte val;
  word adr;
  int len, csum;
  char name[LINESZ], *t;

/*
S4LLAAAAAAAANNNNNNNN,AAAAAAAANNNNNN,AAAAAAAANNNNNNNNNN,CC
LL=length AAAAAAAA=addr NNNN,=name CC=checksum
*/

  p += 2;
  p = scan_byte (&val, p);
  len = val;
  for (; len > 2;)
    {
      p = scan_byte (&val, p);
      adr = (val & 0xff) << 24;
      p = scan_byte (&val, p);
      adr |= (val & 0xff) << 16;
      p = scan_byte (&val, p);
      adr |= (val & 0xff) << 8;
      p = scan_byte (&val, p);
      adr |= (val & 0xff) << 0;
      len -= 8;

      t = strchr (p, ',');
      if (t == 0)
	return;
      strncpy (name, p, t - p);
      len -= t - p;

      if (!newsym (name, adr))
	printf ("out of memory\n");

      p = t + 1;
    }
  p = scan_byte (&val, p);
  csum = val;
}

endrec (p)
     char *p;
{
  word adr;
  char tmp[10];

  switch (p[1])
    {
    case '7':
      strncpy (tmp, &p[4], 8);
      break;
    case '8':
      strncpy (tmp, &p[4], 6);
      break;
    case '9':
      strncpy (tmp, &p[4], 4);
      break;
    }
  atob (&adr, tmp, 16);
  Epc = adr;
  printf ("Entry address is %08x\n", adr);
}


fload (fd, dlproto)
     int fd, dlproto;
{
  int n, i, j, c, len;
  char adat[4], name[LINESZ];
  unsigned char buf[3];
  int bdat, zcnt;
  unsigned long addr;

  getrec (fd, adat);
  if (dlproto == XONXOFF)
    writec (fd, CNTRL (s));
  len = 0;
  if (adat[0] == '/')
    {
      switch (adat[1])
	{
	case 'Z':
	  zcnt = (a2b (adat[2]) << 6) + a2b (adat[3]);
	  buf[0] = buf[1] = buf[2] = 0;
	  for (i = 0; i < zcnt; i++)
	    {
	      memwrite (buf, 3);
	      len += 3;
	    }
	  break;
	case 'B':
	  buf[0] = (a2b (adat[2]) << 6) + a2b (adat[3]);
	  memwrite (buf, 1);
	  len++;
	  break;
	case 'A':
	  flushbuf ();
	  addr = (a2b (adat[2]) << 6) + a2b (adat[3]);
	  addr << = 12;
	  getrec (fd, adat);
	  addr += (a2b (adat[0]) << 6) + a2b (adat[1]);
	  addr << = 12;
	  addr += (a2b (adat[2]) << 6) + a2b (adat[3]);
	  nextaddr = addr;
	  break;
	case 'E':
	  flushbuf ();
	  Epc = nextaddr;
	  printf ("Entry Address  = %08x\n", nextaddr);
	  return (-20);
	  break;
	case 'S':
	  i = 2;
	  j = 0;
	  for (;;)
	    {
	      c = adat[i++];
	      if (c == ',')
		break;
	      name[j++] = c;
	      if (i >= 4)
		{
		  getrec (fd, adat);
		  i = 0;
		}
	    }
	  name[j] = 0;
	  if (!newsym (name, nextaddr))
	    printf ("out of memory\n");
	  break;
	default:
	  return (-1);
	}
    }
  else
    {
      bdat = 0;
      for (i = 0; i < 4; i++)
	{
	  bdat << = 6;
	  bdat += a2b (adat[i]);
	}
      for (i = 2; i >= 0; i--)
	{
	  buf[i] = bdat & MSK;
	  bdat >> = 8;
	}
      memwrite (buf, 3);
      len += 3;
    }
  return (len);
}

getrec (fd, p)
     int fd;
     char *p;
{
  int i;

  if (blkinx >= blksz)
    {
      blksz = read (fd, curblk, BLKSIZE - 1);
      blkinx = 0;
    }
  for (i = 0; i < 4; i++)
    {
      p[i] = curblk[blkinx++];
    }
}

memwrite (dat, n)
     unsigned char *dat;
     int n;
{
  int i;

  for (i = 0; i < n; i++)
    {
      wordbuf << = 8;
      wordbuf += dat[i];
      bytecnt++;
      if (bytecnt >= 4)
	{
	  save_word (nextaddr, wordbuf);
	  nextaddr += 4;
	  bytecnt = 0;
	}
    }
}

flushbuf ()
{
  int posn;

  while (bytecnt > 0)
    {
      posn = (bytecnt - 1) * 8;
      save_byte (nextaddr, getfield (wordbuf, 8, posn));
      nextaddr++;
      bytecnt--;
    }
}

int 
a2b (c)
     char c;
{
/* translate base64 (tx form) to binary
A-Z	0..25
a-z	26..51
0-9	52..61
exta	62
extb	63
*/

  if (c >= 'A' && c <= 'Z')
    return (c - 'A');
  if (c >= 'a' && c <= 'z')
    return (c - 'a' + 26);
  if (c >= '0' && c <= '9')
    return (c - '0' + 52);
  if (c == exta)
    return (62);
  if (c == extb)
    return (63);
  printf ("a2b: %02x: Bad input value\n", c);
  return (0);
}
