#include "mips.h"
#include "pmon.h"
#include "stdio.h"
#include "termio.h"
#include "string.h"

   /* 	user bpts,   cont cmd, trace,    trace thru bra               */
word 	Bpa[MAX_BP], Tmpbpa,   Trcbp1a,  Trcbp2a;             /* addr */
BP 	Bpv[MAX_BP], Tmpbpv,   Trcbp1v,  Trcbp2v;    /* value (instr) */

word DBGREG[NREGS];	/* debugger's register value holder */

jmp_buf jmpb;		/* non-local goto jump buffer */
byte line[LINESZ+1];	/* input line */
byte pattern[LINESZ];	/* pattern buffer */
word cur_adr;		/* default address */
word chksum;		/* used for load and dump */

word as(),setbp(),copy(),dump(),fill(),cont(),go(),help(),dis(),modify();
word registers(),search(),trace(),clrbp(),sdump(),load(),iflush();
word dflush(),dohi(),do_sym(),do_ls(),do_set(),transp();
word stty(),show(),debug(),memtst(),call();

Optdesc b_opts[] = {
	{"-i","hw bpt for instr"},
	{"-d","hw bpt for data"},
	{"-k","hw bpt for kernal only"},
	{"-u","hw bpt for user only"},
	{"-r","hw bpt for data read only"},
	{"-w","hw bpt for data write only"},
	{0}};

Optdesc load_opts[] = {
	{"-s","don't clear symbols"},
	{"-b","don't clear breakpoints"},
	{0}};

Optdesc ls_opts[] = {
	{"-l","long listing"},
	{"-n","numeric order"},
	{"-v","compute value"},
	{"-a","lookup address"},
	{0}};

Optdesc debug_opts[] = {
	{"-s","don't set client sp"},
	{"-v","report protocol errors"},
	{"-V","verbose mode"},
	{0}};

Optdesc g_opts[] = {
	{"-s","don't set client sp"},
	{"<adr>","start address"},
	{"<bptadr>","temporary breakpoint"},
	{"-c <args>","args to be passed to client"},
	{0}};

Optdesc stty_opts[] = {
	{"-v","list possible baud rates and terminal types"},
	{"-a","list all settings"},
	{"<baud>","set baud rate"},
	{"<term>","set terminal type"},
	{"sane","set sane settings"},
	{"ixany","allow any char to restart output"},
	{"-ixany","allow only <start> to restart output"},
	{"ixoff","enable tandem mode"},
	{"-ixoff","disable tandem mode"},
	{0}};

Cmd CmdTable[] = {
	{"h","[*|cmd..]",0,"on-line help",help,1,99},
	{"hi","[cnt]",0,"display command history",dohi,1,2},

	{"m","[adr [val|-s str]..]",0,"modify memory",modify,1,99},
	{"r","[reg* [val|field val]]",0,"display/set register",registers,1,4},

	{"d","[adr [cnt]]",0,"display memory",dump,1,3},
	{"l","[adr [cnt]]",0,"list (disassemble) memory",dis,1,3},

	{"copy","from to siz",0,"copy memory",copy,4,4},
	{"fill","from to {val|-s str}..",0,"fill memory",fill,4,99},

	{"search","from to {val|-s str}..",0,"search memory",search,4,99},
	{"tr","",0,"transparent mode",transp,1,1},

	{"g","[-s] [adr [bptadr]] [-c args]",g_opts,"go execute",go,1,99},
	{"c","[bptadr]",0,"continue execution",cont,1,2},

	{"t","",0,"trace (single step)",trace,1,1},
	{"to","",0,"trace (step over)",trace,1,1},

	{"b","[-idkurw] [adr]..",b_opts,"set break point(s)",setbp,1,99},
	{"db","[numb|*]..",0,"delete break point(s)",clrbp,1,99},

	{"load","[-sb][-c cmdstr]",load_opts,"load memory from hostport",
								load,1,5},
	{"dump","adr siz",0,"dump memory to hostport",sdump,3,3},

	{"set","[name [value]]",0,"display/set variable",do_set,1,3},
	{"stty","[device] [opts]",stty_opts,"set terminal options",stty,1,99},

	{"sym","name value",0,"define symbol",do_sym,3,3},
	{"ls","[-ln sym*|-va adr]",ls_opts,"list symbols",do_ls,1,99},

	{"iflush","",0,"flush I-cache",iflush,1,1},
	{"dflush","",0,"flush D-cache",dflush,1,1},

	{"show","",0,"show copying restrictions",show,1,2},
	{"debug","[-svV]",debug_opts,"enter dbx mode",debug,1,4},

	{"mt","[[addr] size]",0,"memory test",memtst,1,3},
	{"call","addr [val|-s str]..",0,"call function",call,2,99},
	{0}};

void rdbg_init(void){
	int i;

	devinit();
	printf("\n");
printf("PMON version 1.1 [");
#ifdef MIPSEB
printf("BE");
#endif
#ifdef MIPSEL
printf("LE");
#endif
#ifdef FLOATINGPT
printf(",FP");
#endif
printf("], Copyright (C) 1991 LSI LOGIC Corp.\n");

#ifndef SABLE
printf("This is free software, and comes with ABSOLUTELY NO WARRANTY,\n"); 
printf("you are welcome to redistribute it under certain conditions;\n");
printf("type `show' for details.\n");
#endif

	printf("\n\n");
	cur_adr = 0;
	for(i = 0; i < MAX_BP; i++)
		Bpa[i] = NO_BP;
	Tmpbpa = NO_BP;
	Trcbp1a = NO_BP;
	Trcbp2a = NO_BP;
	tracing_mode = TRACE_NO;
}

void rdbg()
{
int i;
char prompt[20],tmp[8],*p;

while(1){
	if(setjmp(jmpb)) printf(" break!\n");
	else ioctl(STDIN,SETINTR,jmpb);

	for (i=3;i<OPEN_MAX;i++) close(i);
	ioctl(STDIN,SETSANE,NULL);
	strcpy(prompt,getenv("prompt"));
	if (p = strchr(prompt,'!')) {
		strdchr(p); /* delete the bang */
		sprintf(tmp,"%d",histno);
		stristr(p,tmp);
		}
	printf("%s",prompt);
	get_cmd(line);
	do_cmd(line);
	}
}

do_cmd(p)
char *p;
{
byte *av[MAX_AC];	/* argument holder */
word ac;		/* # of arguments */
char *cmdlist[20],*t,tmp[11];
int i,nc,j,c,rptcmd;

if (!*p || strempty(p)) { /* blank */
	rptcmd = matchenv("rptcmd");
	if (rptcmd) { /* repeat requested */
		t = gethistn(histno-1);
		if (rptcmd == 1) strcpy(p,t); /* all cmds */
		else if (rptcmd == 2) { /* trace only */
			if (wordsz(t) > 10) return;
			getword(tmp,t);
			if (strequ(tmp,"t") || strequ(tmp,"to")) strcpy(p,tmp);
			else return;
			}
		else {
			printf("bad rptcmd value [%s]\n",getenv("rptcmd"));
			return;
			}
		}
	else return;
	}

nc = 0;
cmdlist[nc++] = p;
for (;*p;) {
	c = *p;
	if(c == SQ || c == DQ) {
		p++;
		while(*p && *p != c) ++p;
		if(*p) p++;
		}
	else if (c == ';') {
		*p++ = 0;
		cmdlist[nc++] = p;
		}
	else p++;
	}

for (j=0;j<nc;j++) {
	ac = argvize(av,cmdlist[j]);
	if(ac > 0){
		for (i=0;CmdTable[i].name != 0;i++) 
			if (strequ(CmdTable[i].name,av[0])) break;
		if (CmdTable[i].name != 0) {
			if (ac >= CmdTable[i].minac && ac <= CmdTable[i].maxac)
				(CmdTable[i].func)(ac,av);
			else
				printf("usage: %s %s\n",CmdTable[i].name,
					CmdTable[i].opts);
			}
		else printf("%s: Command not found.\n",av[0]);
		}
	}
}

help(ac,av)
int ac;
char *av[];
{
int i,j,namemax,optsmax,descmax,len;
int ln,siz;

namemax = optsmax = descmax = 0;
for (i=0;CmdTable[i].name != 0;i++) {
	len = strlen(CmdTable[i].name);
	if (len > namemax) namemax = len;
	len = strlen(CmdTable[i].opts);
	if (len > optsmax) optsmax = len;
	len = strlen(CmdTable[i].desc);
	if (len > descmax) descmax = len;
	}

if (!atob(&siz,getenv("moresz"),10)) {
	printf("%s: bad moresz value\n",getenv("moresz"));
	return(1);
	}
ioctl(STDIN,CBREAK,NULL);

ln = siz;
if (ac >= 2) { /* extended help */
	if (strequ(av[1],"*")) { /* all commands */
		for (i=0;CmdTable[i].name != 0;i++) {
			if (prhelp(i,&ln,siz,namemax,optsmax)) break;
			}
		}
	else { /* specific commands */
		for (j=1;j<ac;j++) {
			for (i=0;CmdTable[i].name != 0;i++) {
				if (strequ(CmdTable[i].name,av[j])) break;
				}
			if (CmdTable[i].name) {
				if (prhelp(i,&ln,siz,namemax,optsmax)) break;
				}
			else printf("%s: command not found\n",av[j]);
			}
		}
	}
else { /* general help only */
	for (i=0;CmdTable[i].name != 0;i++) {
		printf("%*s  %-*s",namemax,CmdTable[i].name,descmax,
				CmdTable[i].desc);
		if (i%2 != 0) printf("\n");
		else printf("   ");
		}
	if (i%2 != 0) printf("\n");
	}
return(0);
}

prhelp(n,lnp,siz,namemax,optsmax)
int n,*lnp,siz,namemax,optsmax;
{
char  prnbuf[80];
Optdesc *p;
int i;

sprintf(prnbuf,"%*s  %-*s    %s",namemax,CmdTable[n].name,optsmax,
					CmdTable[n].opts,CmdTable[n].desc);
if (more(prnbuf,lnp,siz)) return(1);

p = CmdTable[n].optdesc;
if (p != 0) {
	for (i=0;p[i].name;i++) {
		sprintf(prnbuf,"%*s  %9s    %s",namemax,"",p[i].name,p[i].desc);
		if (more(prnbuf,lnp,siz)) return(1);
		}
	}
return(0);
}
