/* mips.s - The part of debugger for R2000/R3000
 *
 *				Copyright (C) 1990 p40wg & Masashi Kamata
 *
 * This program is free software WITHOUT ANY WARRANTY.
 * The author or P40wg don't accept responsibility to anyone for any
 * consequences of using it or for whether it serves any particular
 * purpose or works at all.
 * The copy, modification and redistribution is permited under the
 * conditions in a file named COPYING.
 *					mKamata		Nikkei-MIX
 *					pcs35313	ASCII-PCS
 *					NCC00240	NIFTY-Serve
 *					mka@apricot.apricot.juice.or.jp
 *
 */

#include "mips.h"
#include "pmon.h"

	.text

reset_exception:
	j	start
	.align  8

utlb_miss_exception:
	la	k0,_exception
	li	k1,K1BASE
	or	k0,k1
	j	k0
	.align 7

general_exception:
	la	k0,_exception
	li	k1,K1BASE
	or	k0,k1
	j	k0
	.align  9

	.globl util_routines
util_routines:
	.word	read
	.word	write
	.word	open
	.word	close
	.word	ioctl
	.word	printf
	.word	vsprintf
	.word	ttctl

	.word	rdbg
	.word	getenv
	.word	onintr
	.word	flush_cache
#ifdef FLOATINGPT
	.word	cop1
#else
	.word	0
#endif
	.word	_exception
	.word	envvar
	.word	etext
	.word	edata
	.word	ConfigTable

#define STKSIZE 8192
	.comm	stack,STKSIZE
	.comm   clientstk,STKSIZE
	.comm	hndlrtbl,16*4

	.globl start
	.ent start
start:
	# turn on the RAM
	li	t1,CFGREG
	li	t0,0x003cc62f
	sw	t0,(t1)

	li	t1,RTIC		# refresh
	li	t0,37
	sw	t0,(t1)

	li	v0,SR_BEV
	.set noreorder
	mtc0	v0,C0_SR
	mtc0	zero,C0_CAUSE
	.set reorder

	# set up a K1seg stack
	la	sp,stack+STKSIZE
	or	sp,K1BASE

	# flush the icache
	li	a0,ICACHE
	jal	flush_cache

	# flush the dcache
	li	a0,DCACHE
	jal	flush_cache

	# copy .data to RAM
	# src=etext dst=_fdata stop=edata
	la	t0,etext
	la	t1,_fdata
	la	t2,edata

1:	lw	t3,(t0)
	sw	t3,(t1)
	addu	t0,4
	addu	t1,4
	blt	t1,t2,1b

	# clear bss
	la	v0,_fbss
	la	v1,end
1:
	sw	$0,0x0(v0)
	sw	$0,0x4(v0)
	sw	$0,0x8(v0)
	sw	$0,0xc(v0)
	addu	v0,16
	blt	v0,v1,1b

	# copy handler
	la	t1,handler
	la	t2,ehandler
	li	t3,0x80000000
1:	lw	t0,(t1)
	sw	t0,(t3)
	addu	t1,4
	addu	t3,4
	blt	t1,t2,1b

#ifdef LR33000
	# copy handler
	la	t1,handler
	la	t2,ehandler
	li	t3,0x80000040
1:	lw	t0,(t1)
	sw	t0,(t3)
	addu	t1,4
	addu	t3,4
	blt	t1,t2,1b
#endif

	# copy handler
	la	t1,handler
	la	t2,ehandler
	li	t3,0x80000080
1:	lw	t0,(t1)
	sw	t0,(t3)
	addu	t1,4
	addu	t3,4
	blt	t1,t2,1b

	# flush the dcache
	li	a0,DCACHE
	jal	flush_cache

	# flush the icache
	li	a0,ICACHE
	jal	flush_cache

	# ok to use k0seg stack now
	la	sp,stack+STKSIZE

	# set initial client conditions
	la	k0,DBGREG

	li	v0,0x80020000
	sw	v0,R_EPC*4(k0)

	la	v0,clientstk+STKSIZE
	sw	v0,R_SP*4(k0)

	.set noreorder
	mfc0	v0,C0_SR
	nop
	li	t0,~SR_BEV
	and	v0,t0
	mtc0	v0,C0_SR
	sw	v0,R_STATUS*4(k0)

	mfc0	v0,C0_CAUSE
	nop
	sw	v0,R_CAUSE*4(k0)
	.set reorder

	la	t0,histinit
	jal	t0

	la	t0,rdbg_init
	jal	t0

	la	t0,envinit
	jal	t0

	la	t0,rdbg
	jal	t0

	j	start
	.end start

	.globl handler
	.ent handler
handler:
	.set noat
	la	k0,_exception
	j	k0
ehandler:
	.set at
	.end handler


	.globl _exit
	.ent _exit
_exit:	
#ifndef SABLE 
	break 0
#endif
	j	ra
	.end _exit

	.globl	_go
	.ent _go
_go:
	la	k0,DBGREG
	# restored later		/* $1  (at) */
	lw	v0,R_V0*4(k0)		/* $2  (v0) */
	lw	v1,R_V1*4(k0)		/* $3  (v1) */
	lw	a0,R_A0*4(k0)		/* $4  (a0) */
	lw	a1,R_A1*4(k0)		/* $5  (a1) */
	lw	a2,R_A2*4(k0)		/* $6  (a2) */
	lw	a3,R_A3*4(k0)		/* $7  (a3) */
	lw	t0,R_T0*4(k0)		/* $8  (t0) */
	lw	t1,R_T1*4(k0)		/* $9  (t1) */
	lw	t2,R_T2*4(k0)		/* $10 (t2) */
	lw	t3,R_T3*4(k0)		/* $11 (t3) */
	lw	t4,R_T4*4(k0)		/* $12 (t4) */
	lw	t5,R_T5*4(k0)		/* $13 (t5) */
	lw	t6,R_T6*4(k0)		/* $14 (t6) */
	lw	t7,R_T7*4(k0)		/* $15 (t7) */
	lw	s0,R_S0*4(k0)		/* $16 (s0) */
	lw	s1,R_S1*4(k0)		/* $17 (s1) */
	lw	s2,R_S2*4(k0)		/* $18 (s2) */
	lw	s3,R_S3*4(k0)		/* $19 (s3) */
	lw	s4,R_S4*4(k0)		/* $20 (s4) */
	lw	s5,R_S5*4(k0)		/* $21 (s5) */
	lw	s6,R_S6*4(k0)		/* $22 (s6) */
	lw	s7,R_S7*4(k0)		/* $23 (s7) */
	lw	t8,R_T8*4(k0)		/* $24 (t8) */
	lw	t9,R_T9*4(k0)		/* $25 (t9) */
					/* $26 (k0) */
					/* $27 (k1) */
	lw	gp,R_GP*4(k0)		/* $28 (gp) */
	lw	sp,R_SP*4(k0)		/* $29 (sp) */
	lw	fp,R_FP*4(k0)		/* $30 (fp) */
	lw	ra,R_RA*4(k0)		/* $31 (ra) */
	lw	k1,R_HI*4(k0)		/* Hi */
	mthi	k1
	lw	k1,R_LO*4(k0)		/* Lo */
	mtlo	k1
	.set noreorder
#ifdef LR33000
	lw	k1,R_DCIC*4(k0)		/* DCIC */
	nop
	mtc0	k1,C0_DCIC
	lw	k1,R_BPC*4(k0)		/* BPC */
	nop
	mtc0	k1,C0_BPC
	lw	k1,R_BDA*4(k0)		/* BDA */
	nop
	mtc0	k1,C0_BDA
#else
	lw	k1,R_ENTRYHI*4(k0)	/* EntryHi */
	nop
	mtc0	k1,C0_TLBHI
	lw	k1,R_ENTRYLO*4(k0)	/* EntryLo */
	nop
	mtc0	k1,C0_TLBLO
	lw	k1,R_INDEX*4(k0)	/* Index */
	nop
	mtc0	k1,C0_INX
	lw	k1,R_CONTEXT*4(k0)	/* Context */
	nop
	mtc0	k1,C0_CTEXT
#endif
	.set noat
	lw	k1,R_STATUS*4(k0)	/* Status */
	li	AT,~(SR_KUC|SR_IEC)
	and	k1,k1,AT
	mtc0	k1,C0_SR
	lw	k1,R_CAUSE*4(k0)	/* Cause - For Sw field only */
	nop
	mtc0	k1,C0_CAUSE
	lw	k1,R_K1*4(k0)		/* $27 (k1) */
	lw	AT,R_AT*4(k0)		/* $1  (AT) */
	lw	k0,R_EPC*4(k0)		/* EPC */
	nop
	j	k0
	rfe
	.set at
	.set reorder
	.end _go

	.globl	_exception
	.ent _exception
_exception:
	.set noat
	# see if we have a user defined handler
	.set noreorder
	mfc0	k0,C0_CAUSE
	nop
	.set reorder
	and	k0,CAUSE_EXCMASK
	la	k1,hndlrtbl
	addu	k1,k0
	lw	k0,(k1)
	beq	k0,zero,1f
	la	k1,1f
	j	k0		# jump to user handler

#ifdef FLOATINGPT
1:	# see if it's a cp1 unusable
	.set noreorder
	mfc0	k0,C0_CAUSE
	nop
	.set reorder
	and	k1,k0,CAUSE_EXCMASK
	subu	k1,EXC_CPU	# coprocessor unusable
	bne	k1,zero,1f

	li	k1,CAUSE_CEMASK
	and	k1,k0,k1
	srl	k1,CAUSE_CESHIFT
	subu	k1,1		# coprocessor 1
	bne	k1,zero,1f
	
	la	k1,cop1		# jump to emulation package
	j	k1
#endif
	
1:	la	k0,DBGREG
	sw	AT,R_AT*4(k0)		/* $1  (AT) */
	.set at
	sw	v0,R_V0*4(k0)		/* $2  (v0) */
	sw	v1,R_V1*4(k0)		/* $3  (v1) */
	sw	a0,R_A0*4(k0)		/* $4  (a0) */
	sw	a1,R_A1*4(k0)		/* $5  (a1) */
	sw	a2,R_A2*4(k0)		/* $6  (a2) */
	sw	a3,R_A3*4(k0)		/* $7  (a3) */
	sw	t0,R_T0*4(k0)		/* $8  (t0) */
	sw	t1,R_T1*4(k0)		/* $9  (t1) */
	sw	t2,R_T2*4(k0)		/* $10 (t2) */
	sw	t3,R_T3*4(k0)		/* $11 (t3) */
	sw	t4,R_T4*4(k0)		/* $12 (t4) */
	sw	t5,R_T5*4(k0)		/* $13 (t5) */
	sw	t6,R_T6*4(k0)		/* $14 (t6) */
	sw	t7,R_T7*4(k0)		/* $15 (t7) */
	sw	s0,R_S0*4(k0)		/* $16 (s0) */
	sw	s1,R_S1*4(k0)		/* $17 (s1) */
	sw	s2,R_S2*4(k0)		/* $18 (s2) */
	sw	s3,R_S3*4(k0)		/* $19 (s3) */
	sw	s4,R_S4*4(k0)		/* $20 (s4) */
	sw	s5,R_S5*4(k0)		/* $21 (s5) */
	sw	s6,R_S6*4(k0)		/* $22 (s6) */
	sw	s7,R_S7*4(k0)		/* $23 (s7) */
	sw	t8,R_T8*4(k0)		/* $24 (t8) */
	sw	t9,R_T9*4(k0)		/* $25 (t9) */
					/* $26 (k0) */
	sw	k1,R_K1*4(k0)		/* $27 (k1) */
	sw	gp,R_GP*4(k0)		/* $28 (gp) */
	sw	sp,R_SP*4(k0)		/* $29 (sp) */
	sw	fp,R_FP*4(k0)		/* $30 (fp) */
	sw	ra,R_RA*4(k0)		/* $31 (ra) */
	mfhi	k1
	sw	k1,R_HI*4(k0)		/* Hi */
	mflo	k1
	sw	k1,R_LO*4(k0)		/* Lo */
	.set noreorder
#ifdef LR33000
	mfc0	k1,C0_DCIC
	nop
	sw	k1,R_DCIC*4(k0)		/* DCIC */
	li	t0,~DCIC_DE		# make sure the DE bit is cleared
	and	k1,t0
	mtc0	k1,C0_DCIC
	mfc0	k1,C0_BPC
	nop
	sw	k1,R_BPC*4(k0)		/* BPC */
	mfc0	k1,C0_BDA
	nop
	sw	k1,R_BDA*4(k0)		/* BDA */
#else
	mfc0	k1,C0_TLBHI
	nop
	sw	k1,R_ENTRYHI*4(k0)	/* EntryHi */
	mfc0	k1,C0_TLBLO
	nop
	sw	k1,R_ENTRYLO*4(k0)	/* EntryLo */
	mfc0	k1,C0_INX
	nop
	sw	k1,R_INDEX*4(k0)	/* Index */
	mfc0	k1,C0_RAND
	nop
	sw	k1,R_RANDOM*4(k0)	/* Random */
	mfc0	k1,C0_CTEXT
	nop
	sw	k1,R_CONTEXT*4(k0)	/* Context */
#endif
	mfc0	k1,C0_SR
	nop
	sw	k1,R_STATUS*4(k0)	/* Status */
	mfc0	k1,C0_CAUSE
	nop
	sw	k1,R_CAUSE*4(k0)	/* Cause */
	mfc0	k1,C0_EPC
	nop
	sw	k1,R_EPC*4(k0)		/* EPC */
	mfc0	k1,C0_BADVADDR
	nop
	sw	k1,R_BADVA*4(k0)	/* BadVA */
	mfc0	k1,C0_PRID
	nop
	sw	k1,R_PRID*4(k0)		/* PRId */
	.set reorder
	la	sp,stack+STKSIZE
 #	la	gp,_gp
	la	t0,exception
	j	t0
	.end _exception

	.globl onintr
	.ent onintr
onintr: # a0=exec a1=addr
	sll	a0,2
	la	t1,hndlrtbl
	addu	t1,a0
	sw	a1,(t1)
	j	ra
	.end onintr

	.globl clienttos
	.ent clienttos
clienttos:
	la	v0,clientstk+STKSIZE
	j	ra
	.end clienttos
