/* pmon.h - The part of debugger for MIPS CPUs
 *
 *				Copyright (C) 1990 p40wg & Masashi Kamata
 *
 * This program is free software WITHOUT ANY WARRANTY.
 * The author or P40wg don't accept responsibility to anyone for any
 * consequences of using it or for whether it serves any particular
 * purpose or works at all.
 * The copy, modification and redistribution is permited under the
 * conditions in a file named COPYING.
 *					mKamata		Nikkei-MIX
 *					pcs35313	ASCII-PCS
 *					NCC00240	NIFTY-Serve
 *					mka@apricot.apricot.juice.or.jp
 *
 * PMON Modified and extended for LSI LOGIC Corp. by Phil Bunce.
 *		Phil Bunce, PO Box 2031, Dept 139, Belmont, CA 94002.
 */

#ifdef LANGUAGE_C
typedef char		byte;		/* 1 byte */
typedef short		half;		/* 2 bytes */
typedef long		word;		/* 4 bytes */
typedef unsigned char	ubyte;
typedef unsigned short	uhalf;
typedef unsigned long	uword;

typedef int Func();
typedef void vFunc();
typedef word wFunc();

#define MAXSYM 12
typedef struct Sym {
	struct Sym *anext;
	struct Sym *nnext;
	char *lname;
	unsigned int value;
	char name[MAXSYM];
	} Sym;

typedef word BP;

typedef struct Demo {
	char *name;
	Func *addr;
	} Demo;

typedef struct Optdesc {
	char *name;
	char *desc;
	} Optdesc;

typedef struct Cmd {
	char *name;
	char *opts;
	Optdesc *optdesc;
	char *desc;
	wFunc *func;
	int minac;
	int maxac;
	} Cmd;

typedef enum {
	RD_RS_RT, RT_RS_IMM, OFF, RS_RT_OFF, RS_OFF, NONE, RT_RD, COFUN,
	RS_RT, TARGET, JALR, RS, RD_RT_SFT, LOAD_STORE, RT_IMM, RD, RD_RT_RS,
	RT_RS_SIMM, RT_SIMM, JR, RT_C0, RT_C1, RT_CN, RT_CC1, LDSTC0, LDSTC1,
	LDSTCN, WORD
} TYPE;

#define RS_(x)		(((x) >> 21) & ((1L <<  5) - 1))
#define RT(x)		(((x) >> 16) & ((1L <<  5) - 1))
#define RD_(x)		(((x) >> 11) & ((1L <<  5) - 1))
#define IMM(x)		(((x) >>  0) & ((1L << 16) - 1))
#define TARGET_(x)	(((x) >>  0) & ((1L << 26) - 1))
#define SHAMT(x)	(((x) >>  6) & ((1L <<  5) - 1))

typedef struct {
	char *str;
	long mask, code;
	TYPE type;
} DISTBL;
extern DISTBL distbl[];
extern word DBGREG[];	/* debugger's register value holder */
extern Demo demo[];
extern int histno;	/* current history number */
#endif

#define CTLC	0x03
#define ACK	0x06
#define BS	0x08
#define LF	0x0a
#define NL	LF
#define CR	0x0d
#define CTLQ	0x11
#define CTLS	0x13
#define CAN	0x18
#define ESC	0x1b
#define SP	0x20
#define SQ	'\''
#define DQ	'"'
#define DEL	0x7f

/* flush_cache types */
#define ICACHE  0
#define DCACHE  1
#define IADDR   2

#ifdef SABLE 
#define TIKRATE 1
#else
#define TIKRATE 1000
#endif


#define MAXREC	550	/* max S-record size */
#define LINESZ	200	
#define MAX_AC	100

#define HINIBBLE(x)	(((x) >> 4) & 0xf)
#define LONIBBLE(x)	(((x) >> 0) & 0xf)
#define HIBYTE(x)	(ubyte)((x) >>  8)
#define LOBYTE(x)	(ubyte)((x) >>  0)
#define HIHALF(x)	(uhalf)((x) >> 16)
#define LOHALF(x)	(uhalf)((x) >>  0)
#define CNTRL(x) ('x'&0x1f)
#define cmpstr(x,y) ((strcmp(x,y) == 0)?1:0)
#define getfield(w,s,p) (((w)&(((1<<s)-1)<<p))>>p)

#define MASK16		0x0000ffff
#define MASK26		0x03ffffff
#define SIGNATURE	0x4572696e

#define loop for (;;)

#ifdef LANGUAGE_C
extern jmp_buf jmpb;
extern byte line[LINESZ+1];
extern byte pattern[LINESZ];
extern word cur_adr;
extern byte tracing_mode;
extern word chksum;		/* used for load and dump */
#endif

#define MAX_BP		32
#define NO_BP		-1L
#define BP_CODE		0x0000000dL		/* BREAK instruction */
#define load_bp(a)	load_word(a)
#define save_bp(a,v)	save_word((a),(v))


#ifdef LANGUAGE_C
/* function prototypes */
void reg_rdbg();
byte load_byte(word adr);
void save_byte(word adr, byte v);
byte *scan_byte(byte *valp, byte *src);
half load_half(word adr);
void save_half(word adr, half v);
word load_word(word adr);
void save_word(word adr, word v);
void _go(void);
int setjmp(jmp_buf jb);
void longjmp(jmp_buf jb, int rv);
word disasm(char *dest,word loc, word val);
word dispmem(char *dest,word loc);
void rdbg();
void _jump(word adr);
char *malloc();
void simm();
void imm();
Sym *findvalue();
char *getenv();
void store_breakpoint(void);
void store_trace_breakpoint(void);
word sfmt_load(byte *dst, byte *src, word *adrp);
char *findbang();
char *gethistn();
#endif

#define Gpr		(&DBGREG[R_ZERO])
#define Hi		DBGREG[R_HI]
#define Lo		DBGREG[R_LO]
#define Epc		DBGREG[R_EPC]
#define Status		DBGREG[R_STATUS]
#define Cause		DBGREG[R_CAUSE]
#define Badva		DBGREG[R_BADVA]
#define Prid		DBGREG[R_PRID]

#ifdef LR33000
#  define Dcic			DBGREG[R_DCIC]
#  define Bpc			DBGREG[R_BPC]
#  define Bda			DBGREG[R_BDA]
#else
#  define Context		DBGREG[R_CONTEXT]
#  define Entryhi		DBGREG[R_ENTRYHI]
#  define Entrylo		DBGREG[R_ENTRYLO]
#  define Index			DBGREG[R_INDEX]
#  define Random		DBGREG[R_RANDOM]
#endif


#define MAX_CPL	(80 + 1)	/* max chars per line */

#define TRACE_NO	0	/* no trace/go */
#define TRACE_TB	1	/* trace & break */
#define TRACE_TG	2	/* trace & go */
#define TRACE_GB	3	/* go & break */
#define TRACE_DC	4	/* debug mode continue */
#define TRACE_DS	5	/* debug mode step */

#ifdef LANGUAGE_C
/* external data declarations */
/* extern word DBGREG[]; */
extern word Bpa[MAX_BP], Tmpbpa, Trcbp1a, Trcbp2a;
extern BP Bpv[MAX_BP], Tmpbpv, Trcbp1v, Trcbp2v;
extern char buffer[MAX_CPL];
extern char *bp;
extern long inst, cur_loc;
extern DISTBL *get_distbl();
extern DISTBL distbl[];
#endif

#define htouw(s, p)	htow((s), (word *)(p))

#define load_byte(adr) (*(byte *)adr)
#define load_half(adr) (*(half *)adr)
#define load_word(adr) (*(word *)adr)

/* Target registers index */
#define	R_ZERO		0
#define	R_AT		1
#define	R_V0		2
#define	R_V1		3
#define	R_A0		4
#define	R_A1		5
#define	R_A2		6
#define	R_A3		7
#define	R_T0		8
#define	R_T1		9
#define	R_T2		10
#define	R_T3		11
#define	R_T4		12
#define	R_T5		13
#define	R_T6		14
#define	R_T7		15
#define	R_S0		16
#define	R_S1		17
#define	R_S2		18
#define	R_S3		19
#define	R_S4		20
#define	R_S5		21
#define	R_S6		22
#define	R_S7		23
#define	R_T8		24
#define	R_T9		25
#define	R_K0		26
#define	R_K1		27
#define	R_GP		28
#define	R_SP		29
#define	R_FP		30
#define	R_RA		31
#define R_HI		32
#define R_LO		33
#ifdef LR33000
#  define R_DCIC	34
#  define R_BPC		35
#  define R_BDA		36
#else
#  define R_ENTRYHI	34
#  define R_ENTRYLO	35
#  define R_INDEX	36
#  define R_RANDOM	37
#  define R_CONTEXT	41
#endif
#define R_STATUS	38
#define R_CAUSE		39
#define R_EPC		40
#define R_BADVA		42
#define R_PRID		43
/*
#define R_F0		44
#define R_F1		45
#define R_F2		46
#define R_F3		47
#define R_F4		48
#define R_F5		49
#define R_F6		50
#define R_F7		51
#define R_F8		52
#define R_F9		53
#define R_F10		54
#define R_F11		55
#define R_F12		56
#define R_F13		57
#define R_F14		58
#define R_F15		59
#define R_F16		60
#define R_F17		61
#define R_F18		62
#define R_F19		63
#define R_F20		64
#define R_F21		65
#define R_F22		66
#define R_F23		67
#define R_F24		68
#define R_F25		69
#define R_F26		70
#define R_F27		71
#define R_F28		72
#define R_F29		73
#define R_F30		74
#define R_F31		75
#define R_FCR		76
#define R_FIPR		77
*/
#define NREGS		44

