#include "string.h"

#ifndef TEST
#include "mips.h"
#include "pmon.h"
#endif

#define loop for (;;)
#define BADREG "error: arg2 [%s] bad register name\n"
#define BADFIELD "error: arg3: [%s] bad field name\n"
#define BADBASE "error: arg%d: [%s] base %d value expected\n"
#define BADVALUE "error: arg%d: [%s] max value %d expected\n"
#define BADSVALUE "error: arg4: [%s] bad symbolic value\n"
#define REG_RO "Register is readonly\n"

typedef struct RegSpec {
	int size;
	int lsb;
	char *name;
	int base;
	char **values;
	int ro;
	} RegSpec;

typedef struct RegList {
	word *reg;
	RegSpec *spec;
	char *name;
	char *aname;
	byte ro;
	byte type;
	} RegList;

char *bsvalues[] = {"2","4","8","16",0};

RegSpec Configreg[] = {
	{1,23,"ICD",2,0,0},
	{1,22,"DCD",2,0,0},
	{2,20,"IBS",0,bsvalues,0},
	{2,18,"DBS",0,bsvalues,0},
	{1,17,"IW",2,0,0},
	{4,13,"IOWAIT",10,0,0},
	{1,12,"PW",2,0,0},
	{4,8,"PWAIT",10,0,0},
	{1,7,"CS",2,0,0},
	{1,6,"PE",2,0,0},
	{1,5,"DGE",2,0,0},
	{1,4,"BFD",2,0,0},
	{1,3,"RGN",2,0,0},
	{1,2,"RPC",2,0,0},
	{1,1,"CL",2,0,0},
	{1,0,"DCE",2,0,0},
	{0}};

RegSpec TimerCnl[] = {
	{1,2,"CE",2,0,0},
	{1,1,"IE",2,0,0},
	{1,0,"INT",2,0,0},
	{0}};

RegSpec Stat[] = {
	{4,28,"CU",2,0,0},
	{1,22,"BEV",2,0,0},
	{1,21,"TS",2,0,1},
	{1,20,"PE",2,0,1},
	{1,19,"CM",2,0,0},
	{1,18,"PZ",2,0,0},
	{1,17,"SWC",2,0,0},
	{1,16,"ISC",2,0,0},
	{8,8,"IM&SW",2,0,0},
	{1,5,"KUo",2,0,0},
	{1,4,"IEo",2,0,0},
	{1,3,"KUp",2,0,0},
	{1,2,"IEp",2,0,0},
	{1,1,"KUc",2,0,0},
	{1,0,"IEc",2,0,0},
	{0}};

char *excodes[] = {
	"Int","MOD","TLBL","TLBS","AdEL","AdES","IBE","DBE",
	"Sys","Bp","RI","CpU","Ovf","Resv","Resv","Resv",0};

RegSpec Caus[] = {
	{1,31,"BD",2,0,1},
	{2,28,"CE",10,0,1},
	{6,10,"IP",2,0,1},
	{2,8,"SW",2,0,0},
	{4,2,"EXCODE",0,excodes,1},
	{0}};

RegSpec PRid[] = {
	{8,8,"IMP",10,0,1},
	{8,0,"Rev",10,0,1},
	{0}};

RegSpec DCIC[] = {
	{1,31,"TR",2,0,0},
	{1,30,"UD",2,0,0},
	{1,29,"KD",2,0,0},
	{1,28,"TE",2,0,0},
	{1,27,"DW",2,0,0},
	{1,26,"DR",2,0,0},
	{1,25,"DAE",2,0,0},
	{1,24,"PCE",2,0,0},
	{1,23,"DE",2,0,0},
	{1,13,"D",2,0,0},
	{1,12,"I",2,0,0},
	{1,5,"T",2,0,0},
	{1,4,"W",2,0,0},
	{1,3,"R",2,0,0},
	{1,2,"DA",2,0,0},
	{1,1,"PC",2,0,0},
	{1,0,"DG",2,0,0},
	{0}};

RegList reglist[] = {
	{&Gpr[0],0,"zero","0",1,0},
	{&Gpr[1],0,"at","1",0,0},
	{&Gpr[2],0,"v0","2",0,0},
	{&Gpr[3],0,"v1","3",0,0},
	{&Gpr[4],0,"a0","4",0,0},
	{&Gpr[5],0,"a1","5",0,0},
	{&Gpr[6],0,"a2","6",0,0},
	{&Gpr[7],0,"a3","7",0,0},
	{&Gpr[8],0,"t0","8",0,0},
	{&Gpr[9],0,"t1","9",0,0},
	{&Gpr[10],0,"t2","10",0,0},
	{&Gpr[11],0,"t3","11",0,0},
	{&Gpr[12],0,"t4","12",0,0},
	{&Gpr[13],0,"t5","13",0,0},
	{&Gpr[14],0,"t6","14",0,0},
	{&Gpr[15],0,"t7","15",0,0},
	{&Gpr[16],0,"s0","16",0,0},
	{&Gpr[17],0,"s1","17",0,0},
	{&Gpr[18],0,"s2","18",0,0},
	{&Gpr[19],0,"s3","19",0,0},
	{&Gpr[20],0,"s4","20",0,0},
	{&Gpr[21],0,"s5","21",0,0},
	{&Gpr[22],0,"s6","22",0,0},
	{&Gpr[23],0,"s7","23",0,0},
	{&Gpr[24],0,"t8","24",0,0},
	{&Gpr[25],0,"t9","25",0,0},
	{&Gpr[26],0,"k0","26",0,0},
	{&Gpr[27],0,"k1","27",0,0},
	{&Gpr[28],0,"gp","28",0,0},
	{&Gpr[29],0,"sp","29",0,0},
	{&Gpr[30],0,"s8","30",0,0},
	{&Gpr[31],0,"ra","31",0,0},
	{&Status,Stat,"C0_SR","SR",0,1},
	{&Cause,Caus,"C0_CAUSE","CAUSE",0,1},
	{&Prid,PRid,"C0_PRID","PRID",0,1},
	{&Epc,0,"C0_EPC","EPC",0,1},
	{&Badva,0,"C0_BADVA","BADVA",1,1},
#ifdef LR33000
	{&Dcic,DCIC,"C0_DCIC","DCIC",0,1},
	{&Bpc,0,"C0_BPC","BPC",0,1},
	{&Bda,0,"C0_BDA","BDA",0,1},
	{(word *)CFGREG,Configreg,"M_CFG","CFG",0,1},
	{(word *)TC1,TimerCnl,"M_TC1","TC1",0,1},
	{(word *)TC2,TimerCnl,"M_TC2","TC2",0,1},
	{(word *)TIC1,0,"M_TIC1","TIC1",0,1},
	{(word *)TIC2,0,"M_TIC2","TIC2",0,1},
	{(word *)RTIC,0,"M_RTIC","RTIC",0,1},
#else
	{&Context,0,"C0_CONTEXT","CONTEXT",0,1},
	{&Entrylo,0,"C0_ENTRYLO","ENTRYLO",0,1},
	{&Entryhi,0,"C0_ENTRYHI","ENTRYHI",0,1},
	{&Index,0,"C0_INDEX","INDEX",0,1},
	{&Random,0,"C0_RANDOM","RANDOM",0,1},
#endif
	{0}};

char *btoa();
extern char *regs_sw[];

#ifdef TEST
int ac;
char *av[5];

main(argc,argv)
int argc;
char *argv[];
{
char buf[80];

for (;;) {
	printf("> ");
	gets(buf);
	ac = argvize(av,buf);
	if (ac == 0) continue;
	if (strequ(av[0],"r")) registers();
	else if (strequ(av[0],"q")) break;
	}
}

#endif

registers(ac,av)
int ac;
char *av[];
{
int i,j,len,lmargin,flag,wild;
unsigned int n;
RegSpec *p;
char buf[80],tmp1[20],tmp2[20];
word *pr;


lmargin = 0;
for (j=0;reglist[j].reg != 0;j++) {
	len = strlen(reglist[j].name);
	if (len > lmargin) lmargin = len;
	}

switch (ac) {
	case 1 : /* displ GP registers */
			disp_Gprs(1);
		break;
	case 2 : /* displ selected register(s) */
		if (strequ(av[1],"*")) { /* all regs */
			disp_Gprs(0);
		printf("C0_EPC=%08x C0_BADADDR=%08x C0_BPC=%08x C0_BDA=%08x\n",
				Epc,Badva,Bpc,Bda);
			printf("M_TIC1=%06x       M_TIC2=%06x   M_RTIC=%03x\n",
				*(word *)TIC1,*(word *)TIC2,*(word *)RTIC);

			for (j=0;reglist[j].reg != 0;j++) {
				if (reglist[j].spec == 0) continue;
				disp_reg(&reglist[j],lmargin);
				}
			}
		else { /* display reg(s) by name */
			flag = 0;
			if (strchr(av[1],'*') || strchr(av[1],'?')) wild = 1;
			else wild = 0;
			strtoupper(av[1]);
			for (j=0;reglist[j].reg != 0;j++) {
				strcpy(tmp1,reglist[j].name);
				if (wild) *tmp2 = 0;
				else strcpy(tmp2,reglist[j].aname);
				strtoupper(tmp1);
				strtoupper(tmp2);
				if (strpat(tmp1,av[1]) || strpat(tmp2,av[1])) {
					disp_reg(&reglist[j],lmargin);
					flag = 1;
					}
				}
			if (flag == 0) printf(BADREG,av[1]);
			}
		break;
	case 3 : /* set entire register */
		if (!get_rsa(&n,av[2])) break;
		flag = 0;
		if (strchr(av[1],'*') || strchr(av[1],'?')) wild = 1;
		else wild = 0;
		strtoupper(av[1]);
		for (j=0;reglist[j].reg != 0;j++) {
			strcpy(tmp1,reglist[j].name);
			if (wild) *tmp2 = 0;
			else strcpy(tmp2,reglist[j].aname);
			strtoupper(tmp1);
			strtoupper(tmp2);
			if (strpat(tmp1,av[1]) || strpat(tmp2,av[1])) {
				if (reglist[j].ro) printf(REG_RO);
				else *reglist[j].reg = n;
				flag = 1;
				}
			}
		if (flag == 0) printf(BADREG,av[1]);
		break;
	case 4 : /* set register field */
		for (j=0;reglist[j].reg != 0;j++) {
			if (striequ(av[1],reglist[j].name)) break;
			if (striequ(av[1],reglist[j].aname)) break;
			}
		if (reglist[j].reg == 0) {
			printf(BADREG,av[1]);
			break;
			}
		/* found the reg, now find the field */
		p = reglist[j].spec;
		for (i=0;p[i].name != 0;i++) {
			if (striequ(av[2],p[i].name)) break;
			}
		if (p[i].name == 0) {
			printf(BADFIELD,av[2]);
			break;
			}
		/* found the field, now find the value */	
		if (p[i].ro) {
			printf(REG_RO);
			break;
			}
		/* first check to see if should be symbolic */
		if (p[i].base == 0) {
			for (n=0;p[i].values[n] != 0;n++) {
				if (striequ(av[3],p[i].values[n])) break;
				}
			if (p[i].values[n] == 0) {
				printf(BADSVALUE,av[3]);
				break;
				}
			}
		else {
			if (!atob(&n,av[3],p[i].base)) {
				printf(BADBASE,4,av[3],p[i].base);
				break;
				}
			if (n >= (1<<p[i].size)) {
				printf(BADVALUE,4,av[3],(1<<p[i].size)-1);
				break;
				}
			}
		setfield(reglist[j].reg,p[i].size,p[i].lsb,n);
#ifdef SETBS
		if (reglist[j].reg == (word *)CFGREG 
			&& strequ(p[i].name,"IBS")) setibs(n);
		if (reglist[j].reg == (word *)CFGREG 
			&& strequ(p[i].name,"DBS")) setdbs(n);
#endif
		break;
	}
return(0);
}

disp_Gprs(n)
int n;
{
int i;

if (n==1) {
	printf("    ");
	for (i=0;i<8;i++) printf(" %~8s",regs_sw[i]);
	printf("\n");
	}
printf(" $0-");
for(i = 0; i < 8; i++) printf(" %08x",Gpr[i]);

if (n==1) {
	printf("\n    ");
	for (i=8;i<16;i++) printf(" %~8s",regs_sw[i]);
	}
printf("\n $8-");
for(i = 8; i < 16; i++) printf(" %08x",Gpr[i]);

if (n==1) {
	printf("\n    ");
	for (i=16;i<24;i++) printf(" %~8s",regs_sw[i]);
	}
printf("\n$16-");
for(i = 16; i < 24; i++) printf(" %08x",Gpr[i]);

if (n==1) {
	printf("\n    ");
	for (i=24;i<32;i++) printf(" %~8s",regs_sw[i]);
	}
printf("\n$24-");
for(i = 24; i < 32; i++) printf(" %08x",Gpr[i]);
printf("\n\n");
}

disp_reg(q,lmargin)
RegList *q;
{
int i,value,len,width;
char buf[80];
RegSpec *p;


if (q->spec == 0) {
	if (q->type == 0) printf("$%s-%s-%08x\n",q->aname,q->name,*q->reg);
	else printf("%s: %08x\n",q->name,*q->reg);
	return;
	}

p = q->spec;
printf("%*s: ",lmargin,q->name);

/* print field names */
for (i=0;p[i].size != 0;i++) {
	if (p[i].name == 0) continue;
	len = strlen(p[i].name);
	if (len > p[i].size) width = len;
	else width = p[i].size;
	printf("%~*s ",width,p[i].name);
	}

printf("\n");
for (i=lmargin+2;i>0;i--) printf(" "); /* +2 for ": " */

/* print field values */
for (i=0;p[i].size != 0;i++) {
	if (p[i].name) len = strlen(p[i].name);
	else len = 0;
	if (len > p[i].size) width = len;
	else width = p[i].size;

	value = getfield(*q->reg,p[i].size,p[i].lsb);
	if (p[i].base == 0) strcpy(buf,p[i].values[value]);
	else {
		btoa(buf,value,p[i].base);
		if (p[i].base != 10) str_fmt(buf,p[i].size,FMT_RJUST0);
		}
	printf("%~*s ",width,buf);
	}
printf("\n");
}


/*
int getfield(src,size,lsb)
unsigned int src;
int size,lsb;
{
unsigned int msk;

msk = ((1<<size)-1)<<lsb;
return((src&msk)>>lsb);
}
*/

setfield(dstp,size,lsb,value)
unsigned int *dstp;
int size,lsb,value;
{
unsigned int msk,v;

msk = ((1<<size)-1)<<lsb;
v = *dstp;
v &= ~msk;
v |= value<<lsb;
*dstp =v;
}

getreg(vp,p)
unsigned long *vp;
char *p;
{
int j;

if (isdigit(*p)) {
	if(!atob(&j,p,10) || j > 31) return(0);
	*vp = Gpr[j];
	}
else if ((j = regno(p)) != -1) *vp = Gpr[j];
else {
	for (j=0;reglist[j].reg != 0;j++) {
		if (striequ(p,reglist[j].name)) break;
		if (striequ(p,reglist[j].aname)) break;
		}
	if (!reglist[j].reg) return(0);
	*vp = *reglist[j].reg;
	}
return(1);
}

regno(p)
char *p;
{
int i;

for (i=0;i<32;i++) {
	if (striequ(p,regs_sw[i])) return(i);
	}
return(-1);
}

