#include "stdio.h"
#include "termio.h"
#include "string.h"

struct brate {
	char *str;
	int value;
	};

struct brate btable[] = {
	{"50",B50},
	{"75",B75},
	{"110",B110},
	{"134",B134},
	{"200",B200},
	{"150",B150},
	{"300",B300},
	{"600",B600},
	{"1200",B1200},
	{"1800",B1800},
	{"2400",B2400},
	{"4800",B4800},
	{"9600",B9600},
	{"19200",B19200},
	{"38400",B38400},
	{0}};

stty(ac,av)
int ac;
char *av[];
{
int fd,i,flag,aflag,vflag,j;
struct termio tbuf;
char buf[16];

fd = STDIN;
flag = aflag = vflag = 0;
for (i=1;i<ac;i++) {
	if (av[i][0] == '-') { /* option */
		if (strequ(av[i],"-ixoff")) {
			flag = 1;
			ioctl(fd,TCGETA,&tbuf);
			tbuf.c_iflag &= ~IXOFF;
			ioctl(fd,TCSETAF,&tbuf);
			}
		else if (strequ(av[i],"-ixany")) {
			flag = 1;
			ioctl(fd,TCGETA,&tbuf);
			tbuf.c_iflag &= ~IXANY;
			ioctl(fd,TCSETAF,&tbuf);
			}
		else for (j=1;av[i][j] != 0;j++) {
			if (av[i][j] == 'a') aflag = 1;
			else if (av[i][j] == 'v') vflag = 1;
			}
		}
	else if (isdigit(av[i][0])) { /* set baud rate */
		flag = 1;
		ioctl(fd,TCGETA,&tbuf);
		for (j=0;btable[j].str != 0;j++) {
			if (strequ(av[i],btable[j].str)) break;
			}
		if (btable[j].str == 0) {
			printf("%s: unrecognized baud rate\n",av[i]);
			continue;
			}
		tbuf.c_cflag &= ~CBAUD;
		tbuf.c_cflag |= btable[j].value;
		ioctl(fd,TCSETAF,&tbuf);
		}
	else if (strequ(av[i],"ixoff")) {
		flag = 1;
		ioctl(fd,TCGETA,&tbuf);
		tbuf.c_iflag |= IXOFF;
		ioctl(fd,TCSETAF,&tbuf);
		}
	else if (strequ(av[i],"ixany")) {
		flag = 1;
		ioctl(fd,TCGETA,&tbuf);
		tbuf.c_iflag |= IXANY;
		ioctl(fd,TCSETAF,&tbuf);
		}
	else if (strequ(av[i],"sane")) {
		flag = 1;
		ioctl(fd,SETSANE);
		}
	else if (strpat(av[i],"tty?")) { /* select device */
		fd = open(av[i],0);
		if (fd == -1) {
			printf("can't open %s\n",av[i]);
			return;
			}
		}
	else { /* set term type */
		flag = 1;
		if (ioctl(fd,SETTERM,av[i]) == -1) {
			printf("term type %s not found\n",av[i]);
			continue;
			}
		}
	}

if (vflag) {
	printf("Baud rates:\n\t");
	for (i=0;btable[i].str != 0;i++) {
		printf("%5s ",btable[i].str);
		if (i == 8) printf("\n\t");
		}
	printf("\n");
	printf("Terminal types:\n");
	for (i=0;(i = ioctl(i,TERMTYPE,buf)) != -1;) {
		if ((i%6) == 0) printf("\n");
		printf("%10s ",buf);
		}
	printf("\n");
	}
	
if (flag) return;

flag = 0;
if (ioctl(fd,GETTERM,buf) != -1) {
	printf("term=%s ",buf);
	flag = 1;
	}
ioctl(fd,TCGETA,&tbuf);
for (i=0;btable[i].str != 0;i++) {
	if (tbuf.c_cflag == btable[i].value) break;
	}
if (btable[i].str != 0) {
	printf("baud=%s",btable[i].str);
	flag = 1;
	}
if (flag) printf("\n");

if (!aflag) return;

printf("%sistrip ",((tbuf.c_iflag&ISTRIP)?"":"-")); 
printf("%sixon ",((tbuf.c_iflag&IXON)?"":"-")); 
printf("%sixany ",((tbuf.c_iflag&IXANY)?"":"-"));
printf("%sixoff ",((tbuf.c_iflag&IXOFF)?"":"-"));
printf("%sicanon ",((tbuf.c_lflag&ICANON)?"":"-"));
printf("%secho ",((tbuf.c_lflag&ECHO)?"":"-"));
printf("%sechoe ",((tbuf.c_lflag&ECHOE)?"":"-"));
printf("%sicrnl ",((tbuf.c_iflag&ICRNL)?"":"-")); 
printf("%sonlcr ",((tbuf.c_oflag&ONLCR)?"":"-"));
printf("\n");

printf("erase=%s ",cc2str(buf,tbuf.c_cc[VERASE]));
printf("stop=%s ",cc2str(buf,tbuf.c_cc[V_STOP]));
printf("start=%s ",cc2str(buf,tbuf.c_cc[V_START]));
printf("eol=%s ",cc2str(buf,tbuf.c_cc[VEOL]));
printf("eol2=%s ",cc2str(buf,tbuf.c_cc[VEOL2]));
printf("vintr=%s ",cc2str(buf,tbuf.c_cc[VINTR]));
printf("\n");
}

