#include <stdio.h>
#include <fcntl.h>
#ifdef MIPS
#  include <sys/termio.h>
#else
#  include <termio.h>
#endif
#include <ctype.h>

#ifdef MIPS
   extern char	_ctype[];
#  define	_N	04	/* Numeral (digit) */
#  define	isdigit(c)	((_ctype + 1)[c] & _N)
#endif

#define ACK 0x06

char *getenv();

struct brate {
	char *str;
	int value;
	};

struct brate btable[] = {
	{"50",B50},
	{"75",B75},
	{"110",B110},
	{"134",B134},
	{"200",B200},
	{"150",B150},
	{"300",B300},
	{"600",B600},
	{"1200",B1200},
	{"1800",B1800},
	{"2400",B2400},
	{"4800",B4800},
	{"9600",B9600},
	{"19200",B19200},
	{"38400",B38400},
	{0}};

main(argc,argv) 
int argc; 
char *argv[];
{
char c,buf[256];
int i,fd,line,len;
struct termio tbuf;
int reqdBrate;
char *device, *Bopt;


Bopt = getenv("DLBAUD");
device = getenv("DLPORT");
for (i=1;i<argc;i++) {
	if (argv[i][0] == '-') {
		switch (argv[i][1]) {
		    	default:
				printf("edown: bad option: %s\n",argv[i]);
				exit(1);
			} 
		} 
	else if (isdigit(argv[i][0])) Bopt = argv[i];
	else device = argv[i];
	} 

if (device == 0) {
	printf("edown: device not specified\n");
	exit(-1);
	}

fd = open(device,O_RDWR);
if (fd == -1) {
	printf("edown: can't open %s\n",device);
	exit(-1);
	}

if (Bopt == 0) Bopt = "9600";
for (i=0;btable[i].str != 0;i++) {
	if (!strcmp(Bopt,btable[i].str)) break;
	}
if (btable[i].str == 0) {
	printf("%s: unrecognized baud rate\n",Bopt);
	exit(-1);
	}
reqdBrate = btable[i].value;

/* set device characteristics */
if (ioctl(fd, TCGETA, &tbuf) == -1) {
	fprintf(stderr,"ioctl(TCGETA) failed\n");
	exit(-1);
	}

tbuf.c_iflag &= ~(BRKINT|IGNPAR);
tbuf.c_lflag &= ~(ISIG|ICANON|ECHO);
tbuf.c_oflag  = 0;
tbuf.c_cflag  &= ~CBAUD;
tbuf.c_cflag  |= reqdBrate;
tbuf.c_cc[VEOF] = 1; /* one char */
tbuf.c_cc[VEOL] = 0; /* no wait */

if (ioctl(fd, TCSETAF, &tbuf) == -1) {
	fprintf(stderr,"ioctl(TCSETAF) failed\n");
	exit(-1);
	}

line = 0;
while (gets(buf)) {
	line++;		/* current line number */
	len = strlen(buf);

	/* write a line */
#ifdef RIPS
	putchar('.');
#endif
	if (write(fd,buf,len) != len) {
		fprintf(stderr,"bad write of line %d to device: ",line);
		exit(-1);
		}

	/* send ETX */
	if (write(fd,"\003",1) != 1) {
		fprintf(stderr,"bad write of ETX to device: ");
		exit(-1);
		}

	/* wait for ACK */
	if (read(fd,&c,1) != 1) {
		fprintf(stderr,"bad read from device: ");
		exit(-1);
		}

	c &= 0x7f;	/* mask out parity */
	if (c != ACK) {
		fprintf(stderr, "Bad acknowledgment line %d \n", line);
		exit(-1);
		}

	/* print status line every 32 lines */
	if ((line & 0x1f) == 0) {
		printf("%5d\r",line);
		fflush(stdout);
		}
	}
printf("Done total line %d\n",line);
}
