#include <stdio.h>

/* 
 * This program converts the output from 'nm' into fast-mode records suitable
 * for downloading symbols to LSI Logic's Prom Monitor.
 *
 * usage: nm -B <obj_file> | genFSrec > <file>
 */

#define DEBUG 0
#define BLKSIZE 200

/* special record types */
#define ZEROS "/Z"		/* record contains a count of zeros,
				   actually it's the number of 24-bit
				   records that are all zero */
#define BYTE "/B"		/* record contains a single byte */
#define CHKSUM "/C"		/* checksum to date */
#define ADDR "/A"		/* double length record containing a 32 bit
				   address */
#define END "/E"		/* end of download */

unsigned long addr;
int blksz = 0;
int chksum = 0;

/* define the other 2 chars needed for base 64 */
char exta = ',';
char extb = '.';

main(argc,argv)
int argc;
char *argv[];
{
char buf[100],name[64];
unsigned long addr;
int len,cnt;

while (gets(buf)) {
	sscanf(buf,"%x %*s %s",&addr,name);
	txaddr(addr);
	strcat(name,",");
	for (;;) {
		len = strlen(name)+2;
		if ((len&3) == 0) break;
		strcat(name,"X");
		}
	blksz += len;
	printf("/S%s",name);
	}
}

send12(type,val)
char *type;
int val;
{
char adat[4];

val &= 0xfff;
adat[0] = type[0];
adat[1] = type[1];
adat[2] = b2a(val>>6);
adat[3] = b2a(val);
send(adat,4);
}

send24(val)
unsigned int val;
{
char adat[4];
int i;

val &= 0xffffff;
for (i=3;i>=0;i--) {
	adat[i] = b2a(val);
	val>>=6;
	}
send(adat,4);
}

send(p,n)
char *p;
int n;
{

fwrite(p,sizeof(p[0]),n,stdout);
blksz += n;
if (blksz >= BLKSIZE) {
	putc('\n',stdout);
	blksz = 0;
	}
}

txchksum()
{

send12(CHKSUM,chksum);
chksum = 0;
}

txeof(addr)
unsigned long addr;
{

/* txchksum(chksum); */
txaddr(addr);
send12(END,0);
putc('\n',stdout);
}

txaddr(addr)
unsigned long addr;
{

send12(ADDR,addr>>24);
send24(addr);
if (DEBUG) printf(" %08x ",addr);
}

clrbss(addr,size)
long addr,size;
{
int i,rem;

txaddr(addr);
if (DEBUG) printf("%08x ",size);
rem = size % 3;
size /= 3;
while (size>4095) {
	send12(ZEROS,4095);
	size -= 4095;
	}
send12(ZEROS,size);
for (i=0;i<rem;i++) send12(BYTE,0);
}

a2b(c)
int c;
{
/* translate base64 (tx form) to binary
A-Z	0..25
a-z	26..51
0-9	52..61
exta	62
extb	63
*/

if (c >= 'A' && c <= 'Z') return(c-'A');
if (c >= 'a' && c <= 'z') return(c-'a'+26);
if (c >= '0' && c <= '9') return(c-'0'+52);
if (c == exta) return(62);
if (c == extb) return(63);
printf("a2a: Bad input value\n");
return(0);
}


b2a(c)
int c;
{
/* translate binary to base64 (tx form) 
0..25		A-Z
26..51		a-z
52..61		0-9
62
63
*/

c &= 0x3f;
if (c <= 25)  return('A'+c);
if (c <= 51) return('a'+c-26);
if (c <= 61) return('0'+c-52); 
if (c == 62) return(exta);
if (c == 63) return(extb);
printf("b2a: bad input value\n");
return(0);
}

