#include <stdio.h>
#include "aout.h"

#define DATASIZE  32	/* data size specified in bytes */
#define ADDRSIZE  4	/* address size specified in bytes 
				2 = S1recs, 
				3 = S2recs, 
				4 = S3 recs 
				*/
#define swap(x,y)	((y==2)?byte_swap(x,y):word_swap(x,y))

#define RECSIZE  (DATASIZE + ADDRSIZE + 1)
unsigned long addr;

int sflag = 1;	/* default: do include symbols */
int debug = 0;	/* debug switch */
int prom = 0;	/* prom version */

main(argc,argv)
int argc;
char *argv[];
{
int i,j,num_files;
FILE *fp;

for (i=1;i<argc;i++) {
	if (argv[i][0] == '-') {
		for (j=1;argv[i][j] != 0;j++) {
			if (argv[i][j] == 's') sflag = 0;
			else if (argv[i][j] == 'd') debug = 1;
			else if (argv[i][j] == 'p') prom = 1;
			else fprintf(stderr,"unrecognized option %c\n",
				argv[i][j]);
			}
		}
	else num_files++;
	}

if (num_files == 0) do_file(stdin);
else {
	for (i=1;i<argc;i++) {
		if (argv[i][0] != '-') {
			fp = fopen(argv[i],"r");
			do_file(fp);
			fclose(fp);
			}
		}
	}
}
	
do_file(fp)
FILE *fp;
{
int leflag;
unsigned long offset;
struct filehdr fhdr;
struct aouthdr ohdr;

fread(&fhdr,sizeof(fhdr),1,fp);

if (fhdr.f_magic == 061001) { /* little endian */
	leflag = 1;
	swap(&fhdr.f_magic,sizeof(fhdr.f_magic));
	swap(&fhdr.f_nscns,sizeof(fhdr.f_nscns));
	swap(&fhdr.f_opthdr,sizeof(fhdr.f_opthdr));
	}
else leflag = 0;

if (debug) printf("fhdrsz=%d magic=0%o nscns=%d ohdrsz=%d\n",
	sizeof(fhdr),fhdr.f_magic,fhdr.f_nscns,fhdr.f_opthdr);

if (fhdr.f_magic != 0540 && fhdr.f_magic != 0542) {
	fprintf(stderr,"0%o Bad magic number\n",fhdr.f_magic);
	exit(1);
	}

fread(&ohdr,sizeof(ohdr),1,fp);

if (leflag) {
	swap(&ohdr.entry,sizeof(ohdr.entry));
	swap(&ohdr.tsize,sizeof(ohdr.tsize));
	swap(&ohdr.text_start,sizeof(ohdr.text_start));
	swap(&ohdr.dsize,sizeof(ohdr.dsize));
	swap(&ohdr.data_start,sizeof(ohdr.data_start));
	}

if (debug) printf("entry=%lx tsize=%lx tbase=%lx dsize=%lx dbase=%lx\n",
	ohdr.entry,ohdr.tsize,ohdr.text_start,ohdr.dsize,ohdr.data_start);

addr = 0;
if (!prom) addr = ohdr.text_start;
offset = ((fhdr.f_nscns*sizeof(struct scnhdr))+sizeof(fhdr)+
	sizeof(ohdr)+15)&~15;
fseek(fp,offset,0);
do_sect(fp,ohdr.tsize);
if (!prom) addr = ohdr.data_start;
do_sect(fp,ohdr.dsize);
do_erec(ohdr.entry);
}

do_sect(infile,scnsize)
FILE *infile;
long scnsize;
{
int j,chksum,n;
long cnt;
unsigned char buf[DATASIZE],byte;

for (cnt=0;cnt<scnsize;cnt+=n) {
	chksum = 0;
	if ((scnsize-cnt) >= DATASIZE) n = fread(buf,1,DATASIZE,infile);
	else n = fread(buf,1,scnsize-cnt,infile);
	printf("S%1d%02X",ADDRSIZE-1,n+ADDRSIZE+1);
	chksum += n+ADDRSIZE+1;
	for (j=ADDRSIZE-1;j>=0;j--) {
		byte = (addr>>(j*8)) & 0xff;
		printf("%02X",byte);
		chksum += byte;
		}
	for (j=0;j<n;j++) {
		printf("%02X",buf[j]);
		chksum += buf[j];
		}
	printf("%02X\n",(~chksum)&0xff);
	addr += n;
	}
}

do_erec(entry)
unsigned long entry;
{
int j,chksum;
unsigned char byte;

if (prom) entry = 0;
chksum = 0;
printf("S%1d%02X",7,5);
chksum += 10;
for (j=ADDRSIZE-1;j>=0;j--) {
	byte = (entry>>(j*8)) & 0xff;
	printf("%02X",byte);
	chksum += byte;
	}
printf("%02X\n",(~chksum)&0xff);
}

word_swap(vp)
unsigned long *vp;
{
unsigned short v1,v2;

v1 = (*vp)>>16;
byte_swap(&v1);
v2 = (*vp)&0xffff;
byte_swap(&v2);
*vp = (v2<<16)+v1;
}

byte_swap(vp)
unsigned short *vp;
{

*vp = ((*vp)<<8)+((*vp)>>8);
}
