#! /bin/csh -f
# pmcc -- PMON compiler interface
# This script assumes that the environment variable LSIPKG has been set.
# 
# Apart from the usual cc options, this script supports:
#	-fast		- generate fast-format download records
#	-syms		- don't add symbols to the download file

set LIB = $LSIPKG/lib/BO
set INC = $LSIPKG/include
set fast = 0
set syms = 1
set ofile = "a.out"
set flags = (-DPMCC -DLR33000)

# This while loop does several jobs:
# 	1. strips off any pmcc defined options
# 	2. O3 & O4 options if specified, must preceed path specs
# 	   so extract them from here, and add them to flags.
# 	3. find the output filename, if specified
# 	4. EL option if specified, must preceed path specs
# 	   so extract it from here, and add it to flags.

@ i = 1
while ($i <= $#argv) 
	if ("$argv[$i]" == "-fast") then
		set fast = 1
		set argv[$i] = ""
	else if ("$argv[$i]" == "-syms") then
		set syms = 0
		set argv[$i] = ""
	else if ("$argv[$i]" =~ -O?) then
		set flags = ($flags $argv[$i])
		set argv[$i] = ""
	else if ("$argv[$i]" == "-EL") then
		set flags = ($flags $argv[$i])
		set argv[$i] = ""
		set LIB = $LSIPKG/lib/LO
	else if ("$argv[$i]" == "-o") then
		@ i++
		set ofile = $argv[$i]
	endif
	@ i++
end

# now invoke the compiler
cc -N -T 80020000 $flags -tr -h$LIB -L$LIB -I -I$INC -B $argv:q -lc
if ($status != 0) exit $status

# if you didn't link, we're done
if (! -e $ofile) exit 0

# now make the appropriate download records
if ($fast == 1) then
	if ($syms == 1) then
		nm -B $ofile | genFSrec > $ofile.fast
		if ($status != 0) exit $status
	else
		echo -n "" > $ofile.fast
	endif
	genFrec $ofile >> $ofile.fast
	if ($status != 0) exit $status
else
	if ($syms == 1) then
		nm -B $ofile | genS4rec > $ofile.srec
		if ($status != 0) exit $status
	else
		echo -n "" > $ofile.srec
	endif
	genSrec $ofile >> $ofile.srec
	if ($status != 0) exit $status
endif
exit 0

